# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import


import matplotlib.pyplot as plt
import numpy as np
import time, timeit
import six
import pandas
import CoolProp as CP

CP.CoolProp.set_debug_level(00)
from matplotlib.backends.backend_pdf import PdfPages

all_solvers = ['PT', 'DmolarT', 'HmolarP', 'PSmolar', 'SmolarT', 'PUmolar', 'DmolarP', 'DmolarHmolar', 'DmolarSmolar', 'DmolarUmolar', 'HmolarSmolar', 'HmolarT', 'TUmolar', 'SmolarUmolar', 'HmolarUmolar']
not_implemented_solvers = ['SmolarUmolar', 'HmolarUmolar', 'TUmolar', 'HmolarT']
no_two_phase_solvers = ['PT']

implemented_solvers = [pair for pair in all_solvers if pair not in not_implemented_solvers]

param_labels = dict(Hmolar='Enthalpy [J/mol]/1000',
                    Smolar='Entropy [J/mol/K]/1000',
                    Umolar='Int. Ener. [J/mol]/1000',
                    T='Temperature [K]',
                    Dmolar='Density [mol/m3]/1000',
                    P='Pressure [Pa]/1000')


def split_pair(pair):
    for key in ['Dmolar', 'Hmolar', 'Smolar', 'P', 'T', 'Umolar']:
        if pair.startswith(key):
            return key, pair.replace(key, '')


def split_pair_xy(pair):
    if pair == 'HmolarP':
        return 'Hmolar', 'P'
    elif pair == 'PSmolar':
        return 'Smolar', 'P'
    elif pair == 'PUmolar':
        return 'Umolar', 'P'
    elif pair == 'PT':
        return 'T', 'P'
    elif pair == 'DmolarT':
        return 'Dmolar', 'T'
    elif pair == 'SmolarT':
        return 'Smolar', 'T'
    elif pair == 'TUmolar':
        return 'Umolar', 'T'
    elif pair == 'HmolarT':
        return 'Hmolar', 'T'
    elif pair == 'DmolarP':
        return 'Dmolar', 'P'
    elif pair == 'DmolarHmolar':
        return 'Dmolar', 'Hmolar'
    elif pair == 'DmolarSmolar':
        return 'Dmolar', 'Smolar'
    elif pair == 'DmolarUmolar':
        return 'Dmolar', 'Umolar'
    elif pair == 'HmolarSmolar':
        return 'Smolar', 'Hmolar'
    elif pair == 'SmolarUmolar':
        return 'Smolar', 'Umolar'
    elif pair == 'HmolarUmolar':
        return 'Hmolar', 'Umolar'
    else:
        raise ValueError(pair)


DEBUG_LEVEL = 1


def myprint(level, *args, **kwargs):
    if level > DEBUG_LEVEL:
        print(*args, **kwargs)


class ConsistencyFigure(object):
    def __init__(self, fluid, figsize=(15, 23), backend='HEOS', additional_skips=[], mole_fractions=None, p_limits_1phase=None, T_limits_1phase=None, NT_1phase=40, Np_1phase=40, NT_2phase=20, NQ_2phase=20):

        self.fluid = fluid
        self.backend = backend
        print('***********************************************************************************')
        print('*************** ' + backend + '::' + fluid + ' ************************')
        print('***********************************************************************************')
        self.fig, self.axes = plt.subplots(nrows=5, ncols=3, figsize=figsize)
        self.pairs = all_solvers
        pairs_generator = iter(self.pairs)

        states = [CP.AbstractState(backend, fluid) for _ in range(3)]
        if mole_fractions is not None:
            for state in states:
                state.set_mole_fractions(mole_fractions)
        self.axes_list = []
        for row in self.axes:
            for ax in row:
                pair = six.next(pairs_generator)
                kwargs = dict(p_limits_1phase=p_limits_1phase, T_limits_1phase=T_limits_1phase, NT_1phase=NT_1phase, Np_1phase=Np_1phase,
                              NT_2phase=NT_2phase, NQ_2phase=NQ_2phase)
                self.axes_list.append(ConsistencyAxis(ax, self, pair, self.fluid, self.backend, *states, **kwargs))
                ax.set_title(pair)

        self.calc_saturation_curves()
        self.plot_saturation_curves()

        self.calc_Tmax_curve()
        self.plot_Tmax_curve()

        self.calc_melting_curve()
        self.plot_melting_curve()

        self.tight_layout()

        self.fig.subplots_adjust(top=0.95)
        self.fig.suptitle('Consistency plots for ' + self.fluid, size=14)

        errors = []
        for i, (ax, pair) in enumerate(zip(self.axes_list, self.pairs)):
            if pair not in not_implemented_solvers and pair not in additional_skips:
                errors.append(ax.consistency_check_singlephase())
                if pair not in no_two_phase_solvers:
                    ax.consistency_check_twophase()
            else:
                ax.cross_out_axis()

        self.errors = pandas.concat(errors, sort=True)

    def calc_saturation_curves(self):
        """
        Calculate all the saturation curves in one shot using the state class to save computational time
        """
        HEOS = CP.AbstractState(self.backend, self.fluid)
        self.dictL, self.dictV = {}, {}
        for Q, dic in zip([0, 1], [self.dictL, self.dictV]):
            rhomolar, smolar, hmolar, T, p, umolar = [], [], [], [], [], []
            for _T in np.logspace(np.log10(HEOS.keyed_output(CP.iT_triple)), np.log10(HEOS.keyed_output(CP.iT_critical)), 500):
                try:
                    HEOS.update(CP.QT_INPUTS, Q, _T)
                    if (HEOS.p() < 0): raise ValueError('P is negative:' + str(HEOS.p()))
                    HEOS.T(), HEOS.p(), HEOS.rhomolar(), HEOS.hmolar(), HEOS.smolar()
                    HEOS.umolar()

                    T.append(HEOS.T())
                    p.append(HEOS.p())
                    rhomolar.append(HEOS.rhomolar())
                    hmolar.append(HEOS.hmolar())
                    smolar.append(HEOS.smolar())
                    umolar.append(HEOS.umolar())
                except ValueError as VE:
                    myprint(1, 'satT error:', VE, '; T:', '{T:0.16g}'.format(T=_T), 'T/Tc:', _T / HEOS.keyed_output(CP.iT_critical))

            dic.update(dict(T=np.array(T),
                            P=np.array(p),
                            Dmolar=np.array(rhomolar),
                            Hmolar=np.array(hmolar),
                            Smolar=np.array(smolar),
                            Umolar=np.array(umolar)))

    def plot_saturation_curves(self):
        for ax in self.axes_list:
            ax.label_axes()
            ax.plot_saturation_curves()

    def calc_Tmax_curve(self):
        HEOS = CP.AbstractState(self.backend, self.fluid)
        rhomolar, smolar, hmolar, T, p, umolar = [], [], [], [], [], []

        for _p in np.logspace(np.log10(HEOS.keyed_output(CP.iP_min) * 1.01), np.log10(HEOS.keyed_output(CP.iP_max)), 300):
            try:
                HEOS.update(CP.PT_INPUTS, _p, HEOS.keyed_output(CP.iT_max))
            except ValueError as VE:
                myprint(1, 'Tmax', _p, VE)
                continue

            try:
                T.append(HEOS.T())
                p.append(HEOS.p())
                rhomolar.append(HEOS.rhomolar())
                hmolar.append(HEOS.hmolar())
                smolar.append(HEOS.smolar())
                umolar.append(HEOS.umolar())
            except ValueError as VE:
                myprint(1, 'Tmax access', VE)

        self.Tmax = dict(T=np.array(T),
                         P=np.array(p),
                         Dmolar=np.array(rhomolar),
                         Hmolar=np.array(hmolar),
                         Smolar=np.array(smolar),
                         Umolar=np.array(umolar))

    def plot_Tmax_curve(self):
        for ax in self.axes_list:
            ax.plot_Tmax_curve()

    def calc_melting_curve(self):
        state = CP.AbstractState('HEOS', self.fluid)
        rhomolar, smolar, hmolar, T, p, umolar = [], [], [], [], [], []

        # Melting line if it has it
        if state.has_melting_line():
            pmelt_min = max(state.melting_line(CP.iP_min, -1, -1), state.keyed_output(CP.iP_triple)) * 1.01
            pmelt_max = min(state.melting_line(CP.iP_max, -1, -1), state.keyed_output(CP.iP_max)) * 0.99

            for _p in np.logspace(np.log10(pmelt_min), np.log10(pmelt_max), 100):
                try:
                    Tm = state.melting_line(CP.iT, CP.iP, _p)
                    state.update(CP.PT_INPUTS, _p, Tm)
                    T.append(state.T())
                    p.append(state.p())
                    rhomolar.append(state.rhomolar())
                    hmolar.append(state.hmolar())
                    smolar.append(state.smolar())
                    umolar.append(state.umolar())
                except ValueError as VE:
                    myprint(1, 'melting', VE)

        self.melt = dict(T=np.array(T),
                         P=np.array(p),
                         Dmolar=np.array(rhomolar),
                         Hmolar=np.array(hmolar),
                         Smolar=np.array(smolar),
                         Umolar=np.array(umolar))

    def plot_melting_curve(self):
        for ax in self.axes_list:
            ax.plot_melting_curve()

    def tight_layout(self):
        self.fig.tight_layout()

    def add_to_pdf(self, pdf):
        """ Add this figure to the pdf instance """
        pdf.savefig(self.fig)

    def savefig(self, fname, **kwargs):
        self.fig.savefig(fname, **kwargs)


class ConsistencyAxis(object):
    def __init__(self, axis, fig, pair, fluid, backend, state1, state2, state3,
                 p_limits_1phase=None, T_limits_1phase=None, NT_1phase=40, Np_1phase=40,
                 NT_2phase=20, NQ_2phase=20
                 ):
        self.ax = axis
        self.fig = fig
        self.pair = pair
        self.fluid = fluid
        self.backend = backend
        self.state = state1
        self.state_PT = state2
        self.state_QT = state3
        self.p_limits_1phase = p_limits_1phase
        self.T_limits_1phase = T_limits_1phase
        self.NT_1phase = NT_1phase
        self.Np_1phase = Np_1phase
        self.NQ_2phase = NQ_2phase
        self.NT_2phase = NT_2phase
        # self.saturation_curves()

    def label_axes(self):
        """ Label the axes for the given pair """
        xparam, yparam = split_pair_xy(self.pair)
        self.ax.set_xlabel(param_labels[xparam])
        self.ax.set_ylabel(param_labels[yparam])

        if xparam in ['P', 'Dmolar']:
            self.ax.set_xscale('log')
        if yparam in ['P', 'Dmolar']:
            self.ax.set_yscale('log')

    def plot_saturation_curves(self):
        xparam, yparam = split_pair_xy(self.pair)
        xL = self.to_axis_units(xparam, self.fig.dictL[xparam])
        yL = self.to_axis_units(yparam, self.fig.dictL[yparam])
        xV = self.to_axis_units(xparam, self.fig.dictV[xparam])
        yV = self.to_axis_units(yparam, self.fig.dictV[yparam])
        self.ax.plot(xL, yL, 'k', lw=1)
        self.ax.plot(xV, yV, 'k', lw=1)

    def plot_Tmax_curve(self):
        xparam, yparam = split_pair_xy(self.pair)
        x = self.to_axis_units(xparam, self.fig.Tmax[xparam])
        y = self.to_axis_units(yparam, self.fig.Tmax[yparam])
        self.ax.plot(x, y, 'r', lw=1)

    def plot_melting_curve(self):
        xparam, yparam = split_pair_xy(self.pair)
        x = self.to_axis_units(xparam, self.fig.melt[xparam])
        y = self.to_axis_units(yparam, self.fig.melt[yparam])
        self.ax.plot(x, y, 'b', lw=1)

    def to_axis_units(self, label, vals):
        """ Convert to the units used in the plot """
        if label in ['Hmolar', 'Smolar', 'Umolar', 'Dmolar', 'P']:
            return vals / 1000
        elif label in ['T']:
            return vals
        else:
            raise ValueError(label)

    def consistency_check_singlephase(self):

        tic = time.time()

        # Update the state given the desired set of inputs
        param1, param2 = split_pair(self.pair)
        key1 = getattr(CP, 'i' + param1)
        key2 = getattr(CP, 'i' + param2)
        pairkey = getattr(CP, self.pair + '_INPUTS')

        # Get the keys and indices and values for the inputs needed
        xparam, yparam = split_pair_xy(self.pair)
        xkey = getattr(CP, 'i' + xparam)
        ykey = getattr(CP, 'i' + yparam)

        data = []

        if self.p_limits_1phase is not None:
            # User-specified limits were provided, use them
            p_min, p_max = self.p_limits_1phase
        else:
            # No user-specified limits were provided, use the defaults
            p_min = self.state.keyed_output(CP.iP_min) * 1.01
            p_max = self.state.keyed_output(CP.iP_max)

        for p in np.logspace(np.log10(p_min), np.log10(p_max), self.Np_1phase):

            if self.T_limits_1phase is None:
                # No user-specified limits were provided, using the defaults
                Tmin = self.state.keyed_output(CP.iT_triple)
                if self.state.has_melting_line():
                    try:
                        pmelt_min = self.state.melting_line(CP.iP_min, -1, -1)
                        if p < pmelt_min:
                            T0 = Tmin
                        else:
                            T0 = self.state.melting_line(CP.iT, CP.iP, p)
                    except Exception as E:
                        T0 = Tmin + 1.1
                        data.append(dict(err=str(E), type="melting", input=p))
                        myprint(1, 'MeltingLine:', E)
                else:
                    T0 = Tmin + 1.1
                Tvec = np.linspace(T0, self.state.keyed_output(CP.iT_max), self.NT_1phase)
            else:
                # Use the provided limits for T
                Tvec = np.linspace(self.T_limits_1phase[0], self.T_limits_1phase[1], self.NT_1phase)

            for T in Tvec:

                try:
                    # Update the state using PT inputs in order to calculate all the remaining inputs
                    self.state_PT.update(CP.PT_INPUTS, p, T)
                except ValueError as VE:
                    data.append(dict(err=str(VE), cls="EXCEPTION", type="update", in1="P", val1=p, in2="T", val2=T))
                    myprint(1, 'consistency', VE)
                    continue

                _exception = False
                tic2 = timeit.default_timer()
                try:
                    val1, val2 = self.state_PT.keyed_output(key1), self.state_PT.keyed_output(key2)
                    self.state.update(pairkey, val1, val2)
                    toc2 = timeit.default_timer()
                except ValueError as VE:
                    data.append(dict(err=str(VE), cls="EXCEPTION", type="update", in1=param1, val1=val1, in2=param2, val2=val2))
                    myprint(1, 'update(1p)', self.pair, 'P', p, 'T', T, 'D', self.state_PT.keyed_output(CP.iDmolar), '{0:18.16g}, {1:18.16g}'.format(self.state_PT.keyed_output(key1), self.state_PT.keyed_output(key2)), VE)
                    _exception = True

                x = self.to_axis_units(xparam, self.state_PT.keyed_output(xkey))
                y = self.to_axis_units(yparam, self.state_PT.keyed_output(ykey))

                if not _exception:
                    # Check the error on the density
                    if abs(self.state_PT.rhomolar() / self.state.rhomolar() - 1) < 1e-3 and abs(self.state_PT.p() / self.state.p() - 1) < 1e-3 and abs(self.state_PT.T() - self.state.T()) < 1e-3:
                        data.append(dict(cls="GOOD", x=x, y=y, elapsed=toc2 - tic2))
                        if 'REFPROP' not in self.backend:
                            if self.state_PT.phase() != self.state.phase():
                                myprint(1, 'bad phase', self.pair, '{0:18.16g}, {1:18.16g}'.format(self.state_PT.keyed_output(key1), self.state_PT.keyed_output(key2)), self.state.phase(), 'instead of', self.state_PT.phase())
                    else:
                        data.append(dict(cls="INCONSISTENT", type="update", in1=param1, val1=val1, in2=param2, val2=val2, x=x, y=y))
                        myprint(1, 'bad', self.pair, '{0:18.16g}, {1:18.16g}'.format(self.state_PT.keyed_output(key1), self.state_PT.keyed_output(key2)), 'T:', self.state_PT.T(), 'Drho:', abs(self.state_PT.rhomolar() / self.state.rhomolar() - 1), abs(self.state_PT.p() / self.state.p() - 1), 'DT:', abs(self.state_PT.T() - self.state.T()))

        toc = time.time()
        df = pandas.DataFrame(data)
        bad = df[df.cls == 'INCONSISTENT']
        good = df[df.cls == 'GOOD']
        slowgood = good[good.elapsed > 0.01]
        excep = df[df.cls == 'EXCEPTION']
        badphase = df[df.cls == 'BAD_PHASE']

        self.ax.plot(bad.x, bad.y, 'r+', ms=3)
        self.ax.plot(good.x, good.y, 'k.', ms=1)
        self.ax.plot(excep.x, excep.y, 'rx', ms=3)
        self.ax.plot(slowgood.x, slowgood.y, 'b*', ms=6)
        self.ax.plot(badphase.x, badphase.y, 'o', ms=3, mfc='none')

        print('1-phase took ' + str(toc - tic) + ' s for ' + self.pair)

        if self.pair == 'HmolarSmolar':
            # plt.plot(good.elapsed)
            # plt.title(self.pair)
            # plt.show()

            good.to_excel('times_water.xlsx')
        return df[df.cls != 'GOOD']

    def consistency_check_twophase(self):

        tic = time.time()
        state = self.state

        try:
            if state.fluid_param_string('pure') == 'false':
                print("Not a pure-fluid, skipping two-phase evaluation")
                return
        except:
            pass

        # Update the state given the desired set of inputs
        param1, param2 = split_pair(self.pair)
        key1 = getattr(CP, 'i' + param1)
        key2 = getattr(CP, 'i' + param2)
        pairkey = getattr(CP, self.pair + '_INPUTS')

        # Get the keys and indices and values for the inputs needed
        xparam, yparam = split_pair_xy(self.pair)
        xkey = getattr(CP, 'i' + xparam)
        ykey = getattr(CP, 'i' + yparam)

        data = []
        for q in np.linspace(0, 1, self.NQ_2phase):

            Tmin = state.keyed_output(CP.iT_triple) + 1

            for T in np.linspace(Tmin, state.keyed_output(CP.iT_critical) - 1, self.NT_2phase):

                try:
                    # Update the state using QT inputs in order to calculate all the remaining inputs
                    self.state_QT.update(CP.QT_INPUTS, q, T)
                except ValueError as VE:
                    data.append(dict(err=str(VE), cls="EXCEPTION", type="update", in1="Q", val1=q, in2="T", val2=T))
                    myprint(1, 'consistency', VE)
                    continue

                _exception = False
                try:
                    val1, val2 = self.state_QT.keyed_output(key1), self.state_QT.keyed_output(key2)
                    state.update(pairkey, val1, val2)
                except ValueError as VE:
                    data.append(dict(err=str(VE), cls="EXCEPTION", type="update", in1=param1, val1=val1, in2=param2, val2=val2))
                    myprint(1, 'update_QT', T, q)
                    myprint(1, 'update', param1, self.state_QT.keyed_output(key1), param2, self.state_QT.keyed_output(key2), VE)
                    _exception = True

                x = self.to_axis_units(xparam, self.state_QT.keyed_output(xkey))
                y = self.to_axis_units(yparam, self.state_QT.keyed_output(ykey))

                if not _exception:
                    # Check the error on the density
                    if abs(self.state_QT.rhomolar() / self.state.rhomolar() - 1) < 1e-3 and abs(self.state_QT.p() / self.state.p() - 1) < 1e-3 and abs(self.state_QT.T() - self.state.T()) < 1e-3:
                        data.append(dict(cls="GOOD", x=x, y=y))
                        if 'REFPROP' not in self.backend:
                            if self.state_QT.phase() != self.state.phase():
                                myprint(1, 'bad phase (2phase)', self.pair, '{0:18.16g}, {1:18.16g}'.format(self.state_QT.keyed_output(key1), self.state_QT.keyed_output(key2)), self.state.phase(), 'instead of', self.state_QT.phase())

                    else:
                        myprint(1, 'Q', q)
                        myprint(1, 'bad(2phase)', self.pair, '{0:18.16g}, {1:18.16g}'.format(self.state_QT.keyed_output(key1), self.state_QT.keyed_output(key2)), 'pnew:', self.state.p(), 'pold:', self.state_QT.p(), 'Tnew:', self.state.T(), 'T:', self.state_QT.T(), 'Drho:', abs(self.state_QT.rhomolar() / self.state.rhomolar() - 1), 'DP', abs(self.state_QT.p() / self.state.p() - 1), 'DT:', abs(self.state_QT.T() - self.state.T()))
                        data.append(dict(cls="INCONSISTENT", type="update", in1=param1, val1=val1, in2=param2, val2=val2, x=x, y=y))

        toc = time.time()
        df = pandas.DataFrame(data)
        bad = df[df.cls == 'INCONSISTENT']
        good = df[df.cls == 'GOOD']
        excep = df[df.cls == 'EXCEPTION']
        badphase = df[df.cls == 'BAD_PHASE']

        self.ax.plot(bad.x, bad.y, 'r+', ms=3)
        self.ax.plot(good.x, good.y, 'k.', ms=1)
        self.ax.plot(excep.x, excep.y, 'rx', ms=3)
        self.ax.plot(badphase.x, badphase.y, 'o', ms=3, mfc='none')
        print('2-phase took ' + str(toc - tic) + ' s for ' + self.pair)

    def cross_out_axis(self):
        xlims = self.ax.get_xlim()
        ylims = self.ax.get_ylim()
        self.ax.plot([xlims[0], xlims[1]], [ylims[0], ylims[1]], lw=3, c='r')
        self.ax.plot([xlims[0], xlims[1]], [ylims[1], ylims[0]], lw=3, c='r')

        xparam, yparam = split_pair_xy(self.pair)
        x = 0.5 * xlims[0] + 0.5 * xlims[1]
        y = 0.5 * ylims[0] + 0.5 * ylims[1]
        if xparam in ['P', 'Dmolar']:
            x = (xlims[0] * xlims[1])**0.5
        if yparam in ['P', 'Dmolar']:
            y = (ylims[0] * ylims[1])**0.5

        self.ax.text(x, y, 'Not\nImplemented', ha='center', va='center', bbox=dict(fc='white'))


if __name__ == '__main__':
    PVT = PdfPages('Consistency.pdf')
    CP.CoolProp.set_debug_level(0)
    open('timelog.txt', 'w')
    with open('timelog.txt', 'a+', buffering=1) as fp:
        for fluid in ['METHANOL']:  # CP.__fluids__:
            tic = timeit.default_timer()
            skips = ['DmolarHmolar', 'DmolarSmolar', 'DmolarUmolar', 'HmolarSmolar']
            skips = []
            ff = ConsistencyFigure(fluid, backend='HEOS', additional_skips=skips)  # , NT_1phase = 10, Np_1phase = 10, NT_2phase = 100, NQ_2phase = 0)
            ff.errors.to_excel('Errors' + fluid + '.xlsx')
            toc = timeit.default_timer()
            print('Time to build:', toc - tic, 'seconds')
            ff.add_to_pdf(PVT)
            ff.savefig(fluid + '.png')
            ff.savefig(fluid + '.pdf')
            plt.close()
            fp.write('Time to build: {0} seconds for {1}\n'.format(toc - tic, fluid))
            del ff
    PVT.close()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

# Bring some functions into the Plots namespace for code concision,
# but be careful not to clutter the namespace with too many
# classes and functions.

# Plotting objects and functions
from .Plots import PropertyPlot
from .Common import IsoLine

# Cycle calculation and drawing
from .SimpleCycles import StateContainer
from .SimpleCyclesExpansion import SimpleRankineCycle
from .SimpleCyclesCompression import SimpleCompressionCycle

# Old and deprecated objects
from .SimpleCycles import SimpleCycle, TwoStage, EconomizedCycle



# This file is automatically generated by the generate_constants_module.py script in wrappers/Python.
# DO NOT MODIFY THE CONTENTS OF THIS FILE!

cdef extern from "DataStructures.h" namespace "CoolProp":
	ctypedef enum parameters:
		INVALID_PARAMETER
		igas_constant
		imolar_mass
		iacentric_factor
		irhomolar_reducing
		irhomolar_critical
		iT_reducing
		iT_critical
		irhomass_reducing
		irhomass_critical
		iP_critical
		iP_reducing
		iT_triple
		iP_triple
		iT_min
		iT_max
		iP_max
		iP_min
		idipole_moment
		iT
		iP
		iQ
		iTau
		iDelta
		iDmolar
		iHmolar
		iSmolar
		iCpmolar
		iCp0molar
		iCvmolar
		iUmolar
		iGmolar
		iHelmholtzmolar
		iSmolar_residual
		iDmass
		iHmass
		iSmass
		iCpmass
		iCp0mass
		iCvmass
		iUmass
		iGmass
		iHelmholtzmass
		iviscosity
		iconductivity
		isurface_tension
		iPrandtl
		ispeed_sound
		iisothermal_compressibility
		iisobaric_expansion_coefficient
		iisentropic_expansion_coefficient
		ifundamental_derivative_of_gas_dynamics
		ialphar
		idalphar_dtau_constdelta
		idalphar_ddelta_consttau
		ialpha0
		idalpha0_dtau_constdelta
		idalpha0_ddelta_consttau
		iBvirial
		iCvirial
		idBvirial_dT
		idCvirial_dT
		iZ
		iPIP
		ifraction_min
		ifraction_max
		iT_freeze
		iGWP20
		iGWP100
		iGWP500
		iFH
		iHH
		iPH
		iODP
		iPhase
		iundefined_parameter
	ctypedef enum input_pairs:
		INPUT_PAIR_INVALID
		QT_INPUTS
		PQ_INPUTS
		QSmolar_INPUTS
		QSmass_INPUTS
		HmolarQ_INPUTS
		HmassQ_INPUTS
		DmolarQ_INPUTS
		DmassQ_INPUTS
		PT_INPUTS
		DmassT_INPUTS
		DmolarT_INPUTS
		HmolarT_INPUTS
		HmassT_INPUTS
		SmolarT_INPUTS
		SmassT_INPUTS
		TUmolar_INPUTS
		TUmass_INPUTS
		DmassP_INPUTS
		DmolarP_INPUTS
		HmassP_INPUTS
		HmolarP_INPUTS
		PSmass_INPUTS
		PSmolar_INPUTS
		PUmass_INPUTS
		PUmolar_INPUTS
		HmassSmass_INPUTS
		HmolarSmolar_INPUTS
		SmassUmass_INPUTS
		SmolarUmolar_INPUTS
		DmassHmass_INPUTS
		DmolarHmolar_INPUTS
		DmassSmass_INPUTS
		DmolarSmolar_INPUTS
		DmassUmass_INPUTS
		DmolarUmolar_INPUTS
	ctypedef enum fluid_types:
		FLUID_TYPE_PURE
		FLUID_TYPE_PSEUDOPURE
		FLUID_TYPE_REFPROP
		FLUID_TYPE_INCOMPRESSIBLE_LIQUID
		FLUID_TYPE_INCOMPRESSIBLE_SOLUTION
		FLUID_TYPE_UNDEFINED
	ctypedef enum phases:
		iphase_liquid
		iphase_supercritical
		iphase_supercritical_gas
		iphase_supercritical_liquid
		iphase_critical_point
		iphase_gas
		iphase_twophase
		iphase_unknown
		iphase_not_imposed


cdef extern from "Configuration.h":
	ctypedef enum configuration_keys:
		NORMALIZE_GAS_CONSTANTS
		CRITICAL_WITHIN_1UK
		CRITICAL_SPLINES_ENABLED
		SAVE_RAW_TABLES
		ALTERNATIVE_TABLES_DIRECTORY
		ALTERNATIVE_REFPROP_PATH
		ALTERNATIVE_REFPROP_HMX_BNC_PATH
		ALTERNATIVE_REFPROP_LIBRARY_PATH
		REFPROP_DONT_ESTIMATE_INTERACTION_PARAMETERS
		REFPROP_IGNORE_ERROR_ESTIMATED_INTERACTION_PARAMETERS
		REFPROP_USE_GERG
		REFPROP_USE_PENGROBINSON
		MAXIMUM_TABLE_DIRECTORY_SIZE_IN_GB
		DONT_CHECK_PROPERTY_LIMITS
		HENRYS_LAW_TO_GENERATE_VLE_GUESSES
		PHASE_ENVELOPE_STARTING_PRESSURE_PA
		R_U_CODATA
		VTPR_UNIFAC_PATH
		SPINODAL_MINIMUM_DELTA
		OVERWRITE_FLUIDS
		OVERWRITE_DEPARTURE_FUNCTION
		OVERWRITE_BINARY_INTERACTION
		USE_GUESSES_IN_PROPSSI
		ASSUME_CRITICAL_POINT_STABLE
		VTPR_ALWAYS_RELOAD_LIBRARY
		FLOAT_PUNCTUATION



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import


import matplotlib, warnings
import numpy as np


import CoolProp
from CoolProp.CoolProp import PropsSI
from CoolProp.Plots.Common import BasePlot, PropertyDict, SIunits


def SimpleCycle(Ref, Te, Tc, DTsh, DTsc, eta_a, Ts_Ph='Ph', **kwargs):
    """
    This function plots a simple four-component cycle, on the current axis, or that given by the optional parameter *axis*

    Required parameters:

    * Ref : A string for the refrigerant
    * Te : Evap Temperature in K
    * Tc : Condensing Temperature in K
    * DTsh : Evaporator outlet superheat in K
    * DTsc : Condenser outlet subcooling in K
    * eta_a : Adiabatic efficiency of compressor (no units) in range [0,1]

    Optional parameters:

    * Ts_Ph : 'Ts' for a Temperature-Entropy plot, 'Ph' for a Pressure-Enthalpy
    * axis : An axis to use instead of the active axis
    * skipPlot : If True, won't actually plot anything, just print COP

    """
    warnings.warn("This function has been deprecated. Please consider converting it to an object inheriting from \"BaseCycle\".", DeprecationWarning)

    for i in kwargs:
        warnings.warn("This function has been deprecated, your input \"{0}: {1}\" will be ignored".format(i, kwargs[i]), DeprecationWarning)

    from CoolProp.Plots import SimpleCompressionCycle
    cycle = SimpleCompressionCycle(fluid_ref=Ref, graph_type=Ts_Ph)
    cycle.simple_solve_dt(Te, Tc, DTsh, DTsc, eta_a, SI=True)
    print(cycle.COP_cooling(), cycle.COP_heating())


def TwoStage(Ref, Q, Te, Tc, DTsh, DTsc, eta_oi, f_p, Tsat_ic, DTsh_ic, Ts_Ph='Ph', prints=False, skipPlot=False, axis=None, **kwargs):
    """
    This function plots a two-stage cycle, on the current axis, or that given by the optional parameter *axis*

    Required parameters:

    * Ref : Refrigerant [string]
    * Q : Cooling capacity [W]
    * Te : Evap Temperature [K]
    * Tc : Condensing Temperature [K]
    * DTsh : Evaporator outlet superheat [K]
    * DTsc : Condenser outlet subcooling [K]
    * eta_oi : Adiabatic efficiency of compressor (no units) in range [0,1]
    * f_p : fraction of compressor power lost as ambient heat transfer in range [0,1]
    * Tsat_ic : Saturation temperature corresponding to intermediate pressure [K]
    * DTsh_ic : Superheating at outlet of intermediate stage [K]

    Optional parameters:

    * Ts_Ph : 'Ts' for a Temperature-Entropy plot, 'Ph' for a Pressure-Enthalpy
    * prints : True to print out some values
    * axis : An axis to use instead of the active axis
    * skipPlot : If True, won't actually plot anything, just print COP

    """

    warnings.warn("This function has been deprecated. PLease consider converting it to an object inheriting from \"BaseCycle\".", DeprecationWarning)

    T = np.zeros((8))
    h = np.zeros_like(T)
    p = np.zeros_like(T)
    s = np.zeros_like(T)
    rho = np.zeros_like(T)
    T[0] = np.NAN
    s[0] = np.NAN
    T[1] = Te + DTsh
    pe = PropsSI('P', 'T', Te, 'Q', 1.0, Ref)
    pc = PropsSI('P', 'T', Tc, 'Q', 1.0, Ref)
    pic = PropsSI('P', 'T', Tsat_ic, 'Q', 1.0, Ref)
    Tbubble_c = PropsSI('T', 'P', pc, 'Q', 0, Ref)
    Tbubble_e = PropsSI('T', 'P', pe, 'Q', 0, Ref)

    h[1] = PropsSI('H', 'T', T[1], 'P', pe, Ref)
    s[1] = PropsSI('S', 'T', T[1], 'P', pe, Ref)
    rho[1] = PropsSI('D', 'T', T[1], 'P', pe, Ref)
    T[5] = Tbubble_c - DTsc
    h[5] = PropsSI('H', 'T', T[5], 'P', pc, Ref)
    s[5] = PropsSI('S', 'T', T[5], 'P', pc, Ref)
    rho[5] = PropsSI('D', 'T', T[5], 'P', pc, Ref)
    mdot = Q / (h[1] - h[5])

    rho1 = PropsSI('D', 'T', T[1], 'P', pe, Ref)
    h2s = PropsSI('H', 'S', s[1], 'P', pic, Ref)
    Wdot1 = mdot * (h2s - h[1]) / eta_oi
    h[2] = h[1] + (1 - f_p) * Wdot1 / mdot
    T[2] = PropsSI('T', 'H', h[2], 'P', pic, Ref)
    s[2] = PropsSI('S', 'T', T[2], 'P', pic, Ref)
    rho[2] = PropsSI('D', 'T', T[2], 'P', pic, Ref)
    T[3] = 288
    p[3] = pic
    h[3] = PropsSI('H', 'T', T[3], 'P', pic, Ref)
    s[3] = PropsSI('S', 'T', T[3], 'P', pic, Ref)
    rho[3] = PropsSI('D', 'T', T[3], 'P', pic, Ref)
    rho3 = PropsSI('D', 'T', T[3], 'P', pic, Ref)
    h4s = PropsSI('H', 'T', s[3], 'P', pc, Ref)
    Wdot2 = mdot * (h4s - h[3]) / eta_oi
    h[4] = h[3] + (1 - f_p) * Wdot2 / mdot
    T[4] = PropsSI('T', 'H', h[4], 'P', pc, Ref)
    s[4] = PropsSI('S', 'T', T[4], 'P', pc, Ref)
    rho[4] = PropsSI('D', 'T', T[4], 'P', pc, Ref)

    sbubble_e = PropsSI('S', 'T', Tbubble_e, 'Q', 0, Ref)
    sbubble_c = PropsSI('S', 'T', Tbubble_c, 'Q', 0, Ref)
    sdew_e = PropsSI('S', 'T', Te, 'Q', 1, Ref)
    sdew_c = PropsSI('S', 'T', Tc, 'Q', 1, Ref)

    hsatL = PropsSI('H', 'T', Tbubble_e, 'Q', 0, Ref)
    hsatV = PropsSI('H', 'T', Te, 'Q', 1, Ref)
    ssatL = PropsSI('S', 'T', Tbubble_e, 'Q', 0, Ref)
    ssatV = PropsSI('S', 'T', Te, 'Q', 1, Ref)
    vsatL = 1 / PropsSI('D', 'T', Tbubble_e, 'Q', 0, Ref)
    vsatV = 1 / PropsSI('D', 'T', Te, 'Q', 1, Ref)
    x = (h[5] - hsatL) / (hsatV - hsatL)
    s[6] = x * ssatV + (1 - x) * ssatL
    T[6] = x * Te + (1 - x) * Tbubble_e
    rho[6] = 1.0 / (x * vsatV + (1 - x) * vsatL)

    h[6] = h[5]
    h[7] = h[1]
    s[7] = s[1]
    T[7] = T[1]
    p = [np.nan, pe, pic, pic, pc, pc, pe, pe]
    COP = Q / (Wdot1 + Wdot2)
    RE = h[1] - h[6]

    if prints == True:
        print('x5:', x)
        print('COP:', COP)
        print('COPH', (Q + Wdot1 + Wdot2) / (Wdot1 + Wdot2))
        print(T[2] - 273.15, T[4] - 273.15, p[2] / p[1], p[4] / p[3])
        print(mdot, mdot * (h[4] - h[5]), pic)
        print('Vdot1', mdot / rho1, 'Vdisp', mdot / rho1 / (3500 / 60.) * 1e6 / 0.7)
        print('Vdot2', mdot / rho3, 'Vdisp', mdot / rho3 / (3500 / 60.) * 1e6 / 0.7)
        print(mdot * (h[4] - h[5]), Tc - 273.15)
        for i in range(1, len(T) - 1):
            print('%d & %g & %g & %g & %g & %g \\\\' % (i, T[i] - 273.15, p[i], h[i], s[i], rho[i]))
    else:
        print(Tsat_ic, COP)

    if skipPlot == False:
        if axis == None:
            ax = matplotlib.pyplot.gca()
        else:
            ax = axis
        if Ts_Ph in ['ph', 'Ph']:
            ax.plot(h, p)
        elif Ts_Ph in ['Ts', 'ts']:
            s_copy = s.copy()
            T_copy = T.copy()
            for i in range(1, len(s) - 1):
                ax.plot(s[i], T[i], 'bo', mfc='b', mec='b')
                dT = [0, -5, 5, -20, 5, 5, 5]
                ds = [0, 0.05, 0, 0, 0, 0, 0]
                ax.text(s[i] + ds[i], T[i] + dT[i], str(i))

            s = list(s)
            T = list(T)
            s.insert(7, sdew_e)
            T.insert(7, Te)
            s.insert(5, sbubble_c)
            T.insert(5, Tbubble_c)
            s.insert(5, sdew_c)
            T.insert(5, Tc)

            ax.plot(s, T)
            s = s_copy
            T = T_copy
        else:
            raise TypeError('Type of Ts_Ph invalid')
    return COP


def EconomizedCycle(Ref, Qin, Te, Tc, DTsh, DTsc, eta_oi, f_p, Ti, Ts_Ph='Ts', skipPlot=False, axis=None, **kwargs):
    """
    This function plots an economized cycle, on the current axis, or that given by the optional parameter *axis*

    Required parameters:

    * Ref : Refrigerant [string]
    * Qin : Cooling capacity [W]
    * Te : Evap Temperature [K]
    * Tc : Condensing Temperature [K]
    * DTsh : Evaporator outlet superheat [K]
    * DTsc : Condenser outlet subcooling [K]
    * eta_oi : Adiabatic efficiency of compressor (no units) in range [0,1]
    * f_p : fraction of compressor power lost as ambient heat transfer in range [0,1]
    * Ti : Saturation temperature corresponding to intermediate pressure [K]

    Optional parameters:

    * Ts_Ph : 'Ts' for a Temperature-Entropy plot, 'Ph' for a Pressure-Enthalpy
    * axis : An axis to use instead of the active axis
    * skipPlot : If True, won't actually plot anything, just print COP

    """

    warnings.warn("This function has been deprecated. Please consider converting it to an object inheriting from \"BaseCycle\".", DeprecationWarning)
    from scipy.optimize import newton

    m = 1

    T = np.zeros((11))
    h = np.zeros_like(T)
    p = np.zeros_like(T)
    s = np.zeros_like(T)
    rho = np.zeros_like(T)

    T[0] = np.NAN
    s[0] = np.NAN
    T[1] = Te + DTsh
    pe = PropsSI('P', 'T', Te, 'Q', 1.0, Ref)
    pc = PropsSI('P', 'T', Tc, 'Q', 1.0, Ref)
    pi = PropsSI('P', 'T', Ti, 'Q', 1.0, Ref)
    p[1] = pe
    h[1] = PropsSI('H', 'T', T[1], 'P', pe, Ref)
    s[1] = PropsSI('S', 'T', T[1], 'P', pe, Ref)
    rho[1] = PropsSI('D', 'T', T[1], 'P', pe, Ref)
    h2s = PropsSI('H', 'S', s[1], 'P', pi, Ref)
    wdot1 = (h2s - h[1]) / eta_oi
    h[2] = h[1] + (1 - f_p[0]) * wdot1
    p[2] = pi
    # T[2]=T_hp(Ref,h[2],pi,T2s)
    T[2] = PropsSI('T', 'H', h[2], 'P', pi, Ref)

    s[2] = PropsSI('S', 'T', T[2], 'P', pi, Ref)
    rho[2] = PropsSI('D', 'T', T[2], 'P', pi, Ref)

    T[5] = Tc - DTsc
    h[5] = PropsSI('H', 'T', T[5], 'P', pc, Ref)
    s[5] = PropsSI('S', 'T', T[5], 'P', pc, Ref)
    rho[5] = PropsSI('D', 'T', T[5], 'P', pc, Ref)

    p[5] = pc
    p[6] = pi
    h[6] = h[5]

    p[7] = pi
    p[8] = pi
    p[6] = pi
    T[7] = Ti
    h[7] = PropsSI('H', 'T', Ti, 'Q', 1, Ref)
    s[7] = PropsSI('S', 'T', Ti, 'Q', 1, Ref)
    rho[7] = PropsSI('D', 'T', Ti, 'Q', 1, Ref)
    T[8] = Ti
    h[8] = PropsSI('H', 'T', Ti, 'Q', 0, Ref)
    s[8] = PropsSI('S', 'T', Ti, 'Q', 0, Ref)
    rho[8] = PropsSI('D', 'T', Ti, 'Q', 0, Ref)
    x6 = (h[6] - h[8]) / (h[7] - h[8])  # Vapor Quality
    s[6] = s[7] * x6 + s[8] * (1 - x6)
    rho[6] = 1.0 / (x6 / rho[7] + (1 - x6) / rho[8])
    T[6] = Ti

    # Injection mass flow rate
    x = m * (h[6] - h[8]) / (h[7] - h[6])

    p[3] = pi
    h[3] = (m * h[2] + x * h[7]) / (m + x)
    # T[3]=T_hp(Ref,h[3],pi,T[2])
    T[3] = PropsSI('T', 'H', h[3], 'P', pi, Ref)
    s[3] = PropsSI('S', 'T', T[3], 'P', pi, Ref)
    rho[3] = PropsSI('D', 'T', T[3], 'P', pi, Ref)
    T4s = newton(lambda T: PropsSI('S', 'T', T, 'P', pc, Ref) - s[3], T[2] + 30)
    h4s = PropsSI('H', 'T', T4s, 'P', pc, Ref)
    p[4] = pc
    wdot2 = (h4s - h[3]) / eta_oi
    h[4] = h[3] + (1 - f_p[1]) * wdot2
    # T[4]=T_hp(Ref,h[4],pc,T4s)
    T[4] = PropsSI('T', 'H', h[4], 'P', pc, Ref)
    s[4] = PropsSI('S', 'T', T[4], 'P', pc, Ref)
    rho[4] = PropsSI('D', 'T', T[4], 'P', pc, Ref)

    p[9] = pe
    h[9] = h[8]
    T[9] = Te
    hsatL_e = PropsSI('H', 'T', Te, 'Q', 0, Ref)
    hsatV_e = PropsSI('H', 'T', Te, 'Q', 1, Ref)
    ssatL_e = PropsSI('S', 'T', Te, 'Q', 0, Ref)
    ssatV_e = PropsSI('S', 'T', Te, 'Q', 1, Ref)
    vsatL_e = 1 / PropsSI('D', 'T', Te, 'Q', 0, Ref)
    vsatV_e = 1 / PropsSI('D', 'T', Te, 'Q', 1, Ref)
    x9 = (h[9] - hsatL_e) / (hsatV_e - hsatL_e)  # Vapor Quality
    s[9] = ssatV_e * x9 + ssatL_e * (1 - x9)
    rho[9] = 1.0 / (x9 * vsatV_e + (1 - x9) * vsatL_e)

    s[10] = s[1]
    T[10] = T[1]
    h[10] = h[1]
    p[10] = p[1]

    Tbubble_e = Te
    Tbubble_c = Tc
    sbubble_e = PropsSI('S', 'T', Tbubble_e, 'Q', 0, Ref)
    sbubble_c = PropsSI('S', 'T', Tbubble_c, 'Q', 0, Ref)
    sdew_e = PropsSI('S', 'T', Te, 'Q', 1, Ref)
    sdew_c = PropsSI('S', 'T', Tc, 'Q', 1, Ref)

    Wdot1 = m * wdot1
    Wdot2 = (m + x) * wdot2
    if skipPlot == False:
        if axis == None:
            ax = matplotlib.pyplot.gca()
        else:
            ax = axis
        if Ts_Ph in ['ph', 'Ph']:
            ax.plot(h, p)
            ax.set_yscale('log')
        elif Ts_Ph in ['Ts', 'ts']:
            ax.plot(np.r_[s[7], s[3]], np.r_[T[7], T[3]], 'b')
            s_copy = s.copy()
            T_copy = T.copy()
            dT = [0, -5, 5, -12, 5, 12, -12, 0, 0, 0]
            ds = [0, 0.05, 0.05, 0, 0.05, 0, 0.0, 0.05, -0.05, -0.05]
            for i in range(1, len(s) - 1):
                ax.plot(s[i], T[i], 'bo', mfc='b', mec='b')
                ax.text(s[i] + ds[i], T[i] + dT[i], str(i), ha='center', va='center')

            s = list(s)
            T = list(T)
            s.insert(10, sdew_e)
            T.insert(10, Te)
            s.insert(5, sbubble_c)
            T.insert(5, Tbubble_c)
            s.insert(5, sdew_c)
            T.insert(5, Tc)
            ax.plot(s, T, 'b')

            s = s_copy
            T = T_copy
        else:
            raise TypeError('Type of Ts_Ph invalid')

    COP = m * (h[1] - h[9]) / (m * (h[2] - h[1]) + (m + x) * (h[4] - h[3]))
    for i in range(1, len(T) - 1):
        print('%d & %g & %g & %g & %g & %g \\\\' % (i, T[i] - 273.15, p[i], h[i], s[i], rho[i]))
    print(x, m * (h[1] - h[9]), (m * (h[2] - h[1]) + (m + x) * (h[4] - h[3])), COP)
    mdot = Qin / (h[1] - h[9])
    mdot_inj = x * mdot
    print('x9', x9,)
    print('Qcond', (mdot + mdot_inj) * (h[4] - h[5]), 'T4', T[4] - 273.15)
    print(mdot, mdot + mdot_inj)
    f = 3500 / 60.
    eta_v = 0.7
    print('Vdisp1: ', mdot / (rho[1] * f * eta_v) * 1e6, 'cm^3')
    print('Vdisp2: ', (mdot + mdot_inj) / (rho[1] * f * eta_v) * 1e6, 'cm^3')
    return COP

    # class SimpleCycle(object):
    #    """A class that calculates a simple thermodynamic cycle"""
    #    def __init__(self, *args, **kwargs):
    #        object.__init__(self, *args, **kwargs)
    # (states, steps, fluid):

#             Parameters
#         ----------
#         x_type : int, str
#             Either a letter or an integer that specifies the property type for the x-axis
#         y_type : int, str
#             Either a letter or an integer that specifies the property type for the y-axis
#         states : list
#             A collection of state points that follows a fixed scheme defined
#             in the implementing subclass.
#         fluid_ref : str, CoolProp.AbstractState
#             The fluid property provider, either a subclass of CoolProp.AbstractState
#             or a string that can be used to generate a CoolProp.AbstractState instance
#             via :func:`Common.process_fluid_state`.
#         steps : int
#             The number of steps used for going from one state to another
#
#         for more properties, see :class:`CoolProp.Plots.Common.Base2DObject`.

# # See http://stackoverflow.com/questions/1061283/lt-instead-of-cmp
# class ComparableMixin:
#     """A mixin class that implements all comparing mathods except for __lt__"""
#     def __eq__(self, other):
#         return not self<other and not other<self
#     def __ne__(self, other):
#         return self<other or other<self
#     def __gt__(self, other):
#         return other<self
#     def __ge__(self, other):
#         return not self<other
#     def __le__(self, other):
#         return not other<self


class StatePoint(PropertyDict):
    """A simple fixed dimension dict represented by an object with attributes"""

    # Significant digits in SI units
    ROUND_DECIMALS = {
      CoolProp.iDmass: 5,
      CoolProp.iHmass: 5,
      CoolProp.iP: 2,
      CoolProp.iSmass: 5,
      CoolProp.iT: 5,
      CoolProp.iUmass: 5,
      CoolProp.iQ: 5
    }

    def __iter__(self):
        """Make sure we always iterate in the same order"""
        keys = [CoolProp.iDmass, CoolProp.iHmass, CoolProp.iP, CoolProp.iSmass, CoolProp.iT]
        for key in sorted(keys):
            yield key

    def __str__(self):
        return str(self.__dict__)

    def __prop_compare(self, other, typ):
        # TODO
        if self[typ] is None and other[typ] is None: return 0
        elif self[typ] is None and other[typ] is not None: return -1
        elif self[typ] is not None and other[typ] is None: return 1
        else:
            A = np.round(self[typ], self.ROUND_DECIMALS[typ])
            B = np.round(other[typ], self.ROUND_DECIMALS[typ])
            if A > B: return 1
            elif A < B: return -1
            elif A == B: return 0
            else: raise ValueError("Comparison failed.")

    def __eq__(self, other):
        for i in self:
            if not self.__prop_compare(other, i) == 0:
                return False
        return True

    def __hash__(self):
        return hash(repr(self))


class StateContainer(object):
    """A collection of values for the main properties, built to mixin with :class:`CoolProp.Plots.Common.PropertyDict`

    Examples
    --------
    This container has overloaded accessor methods. Just pick your own flavour
    or mix the styles as you like:

    >>> from __future__ import print_function
    >>> import CoolProp
    >>> from CoolProp.Plots.SimpleCycles import StateContainer
    >>> T0 = 300.000; p0 = 200000.000; h0 = 112745.749; s0 = 393.035
    >>> cycle_states = StateContainer()
    >>> cycle_states[0,'H'] = h0
    >>> cycle_states[0]['S'] = s0
    >>> cycle_states[0][CoolProp.iP] = p0
    >>> cycle_states[0,CoolProp.iT] = T0
    >>> cycle_states[1,"T"] = 300.064
    >>> print(cycle_states)
    Stored State Points:
    state        T (K)       p (Pa)    d (kg/m3)     h (J/kg)   s (J/kg/K)
        0      300.000   200000.000            -   112745.749      393.035
        1      300.064            -            -            -            -

    """

    def __init__(self, unit_system=SIunits()):
        self._points = {}
        self._units = unit_system

    @property
    def points(self): return self._points

    @points.setter
    def points(self, value): self._points = value

    @property
    def units(self): return self._units

    @units.setter
    def units(self, value): self._units = value

    def get_point(self, index, SI=True):
        if SI:
            state = self[index]
        else:
            state = self[index]
            for i in state:
                state[i] = self.units[i].from_SI(state[i])
        return state

    def set_point(self, index, value, SI=True):
        if SI:
            self._points[index] = value
        else:
            for i in value:
                self._points[index][i] = self.units[i].to_SI(value[i])

    def _list_like(self, value):
        """Try to detect a list-like structure excluding strings"""
        return (not hasattr(value, "strip") and
            (hasattr(value, "__getitem__") or
            hasattr(value, "__iter__")))
        # return is_sequence(value) # use from pandas.core.common import is_sequence

    def __len__(self):
        """Some cheating to get the correct behaviour"""
        return len(self._points)

    def __iter__(self):
        """Make sure we iterate in the righ order"""
        for key in sorted(self._points):
            yield key

    def __getitem__(self, index):
        """Another tweak that changes the default access path"""
        if self._list_like(index):
            len_var = len(index)
            if len_var == 0:
                raise IndexError("Received empty index.")
            elif len_var == 1:
                return self._points[index[0]]
            elif len_var == 2:
                return self._points[index[0]][index[1]]
            else:
                raise IndexError("Received too long index.")
        return self._points[index]

    def __setitem__(self, index, value):
        """Another tweak that changes the default access path"""
        if self._list_like(index):
            len_var = len(index)
            if len_var == 0:
                raise IndexError("Received empty index.")
            elif len_var == 1:
                self._points[index[0]] = value
            elif len_var == 2:
                # safeguard against empty entries
                if index[0] not in self._points:
                    self._points[index[0]] = StatePoint()
                self._points[index[0]][index[1]] = value
            else:
                raise IndexError("Received too long index.")
        else:
            self._points[index] = value

    def __str__(self):
        out = "Stored State Points:\n"
        keys = True
        for i in self._points:
            if keys:
                row = [u"{0:>5s}".format("state")]
                for j in self._points[i]:
                    label = u"{0:s} ({1:s})".format(self.units[j].symbol, self.units[j].unit)
                    row.append(u"{0:>11s}".format(label))
                out = out + u"  ".join(row) + "\n"
                keys = False
            row = [u"{0:>5s}".format(str(i))]
            for j in self._points[i]:
                try:
                    row.append(u"{0:11.3f}".format(self.units[j].from_SI(self._points[i][j])))
                except:
                    row.append(u"{0:>11s}".format("-"))
            out = out + u"  ".join(row) + "\n"
        return out

    def append(self, new):
        i = 0 + self.__len__()
        for j in new:
            self[i, j] = new[j]
        return self

    def extend(self, new):
        i = 0 + self.__len__()
        for j in new:
            for k in new[j]:
                self[i, k] = new[j][k]
            i = i + 1
        return self

    @property
    def D(self): return np.array([self._points[k].D for k in self])

    @property
    def H(self): return np.array([self._points[k].H for k in self])

    @property
    def P(self): return np.array([self._points[k].P for k in self])

    @property
    def S(self): return np.array([self._points[k].S for k in self])

    @property
    def T(self): return np.array([self._points[k].T for k in self])

    @property
    def U(self): return np.array([self._points[k].U for k in self])

    @property
    def Q(self): return np.array([self._points[k].Q for k in self])


class BaseCycle(BasePlot):
    """A simple thermodynamic cycle, should not be used on its own."""

    # Define the iteration keys
    PROPERTIES = {
      CoolProp.iDmass: 'density',
      CoolProp.iHmass: 'specific enthalpy',
      CoolProp.iP: 'pressure',
      CoolProp.iSmass: 'specific entropy',
      CoolProp.iT: 'temperature'
    }

    STATECOUNT = 0
    """A list of accepted numbers of states"""

    STATECHANGE = None
    """A list of lists of tuples that defines how the state transitions
    behave for the corresponding entry in BaseCycle.STATECOUNT"""

    def __init__(self, fluid_ref, graph_type, unit_system='EUR', **kwargs):
        """Initialises a simple cycle calculator

        Parameters
        ----------
        fluid_ref : str, CoolProp.AbstractState
            The fluid property provider, either a subclass of CoolProp.AbstractState
            or a string that can be used to generate a CoolProp.AbstractState instance
            via :func:`Common.process_fluid_state`.
        graph_type : string
            The graph type to be plotted, like \"PH\" or \"TS\"
        unit_system : string, ['EUR','KSI','SI']
            Select the units used for the plotting.  'EUR' is bar, kJ, C; 'KSI' is kPa, kJ, K; 'SI' is Pa, J, K

        for more properties, see :class:`CoolProp.Plots.Common.BasePlot`.
        """
        self._cycle_states = StateContainer()
        self._steps = 2
        BasePlot.__init__(self, fluid_ref, graph_type, unit_system, **kwargs)

    @property
    def cycle_states(self): return self._cycle_states

    @cycle_states.setter
    def cycle_states(self, value):
        if len(value) != self.STATECOUNT:
            raise ValueError("Your number of states ({0:d}) is not in the list of allowed state counts: {1:s}.".format(len(value), str(self.STATECOUNT)))
        self._cycle_states = value

    @property
    def steps(self): return self._steps

    @steps.setter
    def steps(self, value): self._steps = int(max([value, 2]))

    @BasePlot.system.setter
    def system(self, value):
        if value in self.UNIT_SYSTEMS:
            self._system = self.UNIT_SYSTEMS[value]
        elif isinstance(value, PropertyDict):
            self._system = value
        else:
            raise ValueError("Invalid unit_system input \"{0:s}\", expected a string from {1:s}".format(str(value), str(self.UNIT_SYSTEMS.keys())))
        self._cycle_states.units = self._system

    def valid_states(self):
        """Check the formats of BaseCycle.STATECOUNT and BaseCycle.STATECHANGE"""
        if len(self.STATECHANGE) != self.STATECOUNT:
            raise ValueError("Invalid number of states and or state change operations")
        return True

    def fill_states(self, objs=None):
        """Try to populate all fields in the state objects"""

        if objs is None:
            objs = self._cycle_states
            local = True
        else:
            local = False

        for i in objs:
            full = True
            for j in objs[i]:
                if objs[i][j] is None:
                    full = False
            if full: continue
            if (objs[i][CoolProp.iDmass] is not None and
              objs[i][CoolProp.iT] is not None):
                self._state.update(CoolProp.DmassT_INPUTS, objs[i][CoolProp.iDmass], objs[i][CoolProp.iT])
            elif (objs[i][CoolProp.iP] is not None and
              objs[i][CoolProp.iHmass] is not None):
                self._state.update(CoolProp.HmassP_INPUTS, objs[i][CoolProp.iHmass], objs[i][CoolProp.iP])
            elif (objs[i][CoolProp.iP] is not None and
              objs[i][CoolProp.iSmass] is not None):
                self._state.update(CoolProp.PSmass_INPUTS, objs[i][CoolProp.iP], objs[i][CoolProp.iSmass])
            else:
                warnings.warn("Please fill the state[{0:s}] manually.".format(str(i)))
                continue
            for j in objs[i]:
                if objs[i][j] is None:
                    objs[i][j] = self._state.keyed_output(j)

        if local: self._cycle_states = objs
        return objs

    def state_change(self, in1, in2, start, ty1='lin', ty2='lin'):
        """Calculates a state change defined by the properties in1 and in2

        Uses self.states[start] and self.states[start+1] (or self.states[0]) to define
        the process and interpolates between the values.

        Parameters
        ----------
        in1 : int
            The index of the first defined property.
        in2 : int
            The index of the second defined property.
        start : int
            The index of the start state.
        ty1 : str
            The key that defines the type of state change for in1, lin or log.
        ty2 : str
            The key that defines the type of state change for in2, lin or log.

        Returns
        -------
        scalar or array_like
            a list of the length of self.steps+1 that describes the process. It includes start and end state.
        """
        self.fill_states()
        end = start + 1
        if end >= len(self.cycle_states): end -= len(self.cycle_states)
        start = self.cycle_states[start]
        end = self.cycle_states[end]
        #
        val = []
        inv = [in1, in2]
        typ = [ty1, ty2]
        for i, v in enumerate(inv):
            if typ[i] == 'lin':
                val.append(np.linspace(start[v], end[v], self.steps))
            elif typ[i] == 'log':
                val.append(np.logspace(np.log10(start[v]), np.log10(end[v]), self.steps))
            else:
                raise ValueError("Unknown range generator {0:s}".format(str(typ[i])))

        sc = StateContainer(self._system)
        for i, _ in enumerate(val[0]):
            sc[i, inv[0]] = val[0][i]
            sc[i, inv[1]] = val[1][i]

        return self.fill_states(sc)

    def get_state_change(self, index):
        return self.STATECHANGE[index](self)

    def get_state_changes(self):
        sc = self.get_state_change(0)
        for i in range(1, self.STATECOUNT):
            sc.extend(self.get_state_change(i))
        return sc



print("import: 'CoolProp'")
import CoolProp

print("import: 'CoolProp.tests'")
import CoolProp.tests

print("import: 'CoolProp.Plots'")
import CoolProp.Plots

print("import: 'CoolProp.GUI'")
import CoolProp.GUI




from __future__ import division, print_function
import CoolProp.CoolProp as CP
from CoolProp.State import State


def first_derivative(S, func, iVal, Val, iConstant, Constant, epsilon=1e-3):

    S.update({iVal: Val, iConstant: Constant})
    val1 = func()

    S.update({iVal: Val + epsilon, iConstant: Constant})
    val2 = func()

    S.update({iVal: Val, iConstant: Constant})

    return (val2 - val1) / epsilon


def second_derivative(S, func, iVal, Val, iConstant, Constant, epsilon=2):

    S.update({iVal: Val - epsilon, iConstant: Constant})
    val1 = func()

    S.update({iVal: Val, iConstant: Constant})
    val2 = func()

    S.update({iVal: Val + epsilon, iConstant: Constant})
    val3 = func()

    S.update({iVal: Val, iConstant: Constant})

    print(val1, val2, val3, S.T, S.p, S.rho, (val1 - 2 * val2 + val3))

    return (val1 - 2 * val2 + val3) / (epsilon * epsilon)


def teest_1phase_first_derivatives():

    for US in [CoolProp.UNIT_SYSTEM_SI, CoolProp.UNIT_SYSTEM_KSI]:
        CP.set_standard_unit_system(US)

        S = State('R134a', dict(T=300, D=1))

        l = [(S.get_rho, 'T', S.T, 'P', S.p, S.PFC.drhodT_constp),
             (S.get_rho, 'P', S.p, 'T', S.T, S.PFC.drhodp_constT),
             (S.get_p, 'D', S.rho, 'T', S.T, S.PFC.dpdrho_constT),
             # (S.get_p,'D',S.rho,'H',S.h,S.PFC.dpdrho_consth), #(these inputs not supported)
             (S.get_p, 'T', S.T, 'D', S.rho, S.PFC.dpdT_constrho),
             # (S.get_p,'T',S.T,'H',S.h,S.PFC.dpdT_consth),     #(these inputs not supported)
             (S.get_h, 'D', S.rho, 'T', S.T, S.PFC.dhdrho_constT),
             (S.get_h, 'D', S.rho, 'P', S.p, S.PFC.dhdrho_constp),
             (S.get_h, 'T', S.T, 'D', S.rho, S.PFC.dhdT_constrho),
             (S.get_h, 'T', S.T, 'P', S.p, S.PFC.dhdT_constp),
             (S.get_h, 'P', S.p, 'T', S.T, S.PFC.dhdp_constT),
             (S.get_s, 'D', S.rho, 'T', S.T, S.PFC.dsdrho_constT),
             (S.get_s, 'T', S.T, 'D', S.rho, S.PFC.dsdT_constrho),
             (S.get_s, 'D', S.rho, 'P', S.p, S.PFC.dsdrho_constp),
             (S.get_s, 'T', S.T, 'P', S.p, S.PFC.dsdT_constp),
             (S.get_s, 'P', S.p, 'T', S.T, S.PFC.dsdp_constT),

            ]
        for args in l:
            yield (check_1phase_first_derivatives,) + (S,) + args


def check_1phase_first_derivatives(S, func, iVal, Val, iConstant, Constant, deriv_func):

    Deriv_val = first_derivative(S, func, iVal, Val, iConstant, Constant)
    EOS_val = deriv_func()
    if abs(EOS_val / Deriv_val - 1) > 1e-2:
        raise ValueError('Finite Diff: ' + str(Deriv_val) + ' EOS: ' + str(EOS_val))


def teest_sat_first_derivatives():

    for US in [CoolProp.UNIT_SYSTEM_SI, CoolProp.UNIT_SYSTEM_KSI]:
        CP.set_standard_unit_system(US)

        S = State('R134a', dict(T=300, Q=1))

        l = [(S.get_T, 'P', S.p, 'Q', 0, S.PFC.dTdp_along_sat),
             (S.get_rho, 'P', S.p, 'Q', 0, S.PFC.drhodp_along_sat_liquid),
             (S.get_rho, 'P', S.p, 'Q', 1, S.PFC.drhodp_along_sat_vapor),
             (S.get_rho, 'T', S.T, 'Q', 0, S.PFC.drhodT_along_sat_liquid),
             (S.get_rho, 'T', S.T, 'Q', 1, S.PFC.drhodT_along_sat_vapor),
             (S.get_h, 'P', S.p, 'Q', 0, S.PFC.dhdp_along_sat_liquid),
             (S.get_h, 'P', S.p, 'Q', 1, S.PFC.dhdp_along_sat_vapor),
             (S.get_s, 'P', S.p, 'Q', 0, S.PFC.dsdp_along_sat_liquid),
             (S.get_s, 'P', S.p, 'Q', 1, S.PFC.dsdp_along_sat_vapor),
            ]
        for args in l:
            yield (check_sat_first_derivatives,) + (S,) + args


def check_sat_first_derivatives(S, func, iVal, Val, iConstant, Constant, deriv_func):

    Deriv_val = first_derivative(S, func, iVal, Val, iConstant, Constant)
    EOS_val = deriv_func()
    if abs(EOS_val / Deriv_val - 1) > 1e-2:
        raise ValueError('Finite Diff: ' + str(Deriv_val) + ' EOS: ' + str(EOS_val))


def teest_sat_second_derivatives():
    for US in [CoolProp.UNIT_SYSTEM_SI, CoolProp.UNIT_SYSTEM_KSI]:
        CP.set_standard_unit_system(US)

        S = State('R134a', dict(T=290, Q=1))

        l = [(S.get_T, 'P', S.p, 'Q', 0, S.PFC.d2Tdp2_along_sat),
             (S.get_rho, 'P', S.p, 'Q', 0, S.PFC.d2rhodp2_along_sat_liquid),
             (S.get_rho, 'P', S.p, 'Q', 1, S.PFC.d2rhodp2_along_sat_vapor),
             (S.get_h, 'P', S.p, 'Q', 0, S.PFC.d2hdp2_along_sat_liquid),
             (S.get_h, 'P', S.p, 'Q', 1, S.PFC.d2hdp2_along_sat_vapor),
             (S.get_s, 'P', S.p, 'Q', 0, S.PFC.d2sdp2_along_sat_liquid),
             (S.get_s, 'P', S.p, 'Q', 1, S.PFC.d2sdp2_along_sat_vapor),
            ]
        for args in l:
            yield (check_sat_second_derivatives,) + (S,) + args


def check_sat_second_derivatives(S, func, iVal, Val, iConstant, Constant, deriv_func):

    Deriv_val = second_derivative(S, func, iVal, Val, iConstant, Constant)
    EOS_val = deriv_func()
    if abs(EOS_val / Deriv_val - 1) > 1e-2:
        raise ValueError('Finite Diff: ' + str(Deriv_val) + ' EOS: ' + str(EOS_val))


if __name__ == '__main__':
    import nose
    nose.runmodule()




import numpy as np
import matplotlib.pyplot as plt
from CoolProp.HumidAirProp import HAProps

Tdb = np.linspace(-10, 55, 100) + 273.15

# Make the figure and the axes
fig = plt.figure(figsize=(10, 8))
ax = fig.add_axes((0.15, 0.15, 0.8, 0.8))
ax.set_xlim(Tdb[0] - 273.15, Tdb[-1] - 273.15)
ax.set_ylim(0, 0.03)
ax.set_xlabel(r"Dry bulb temperature [$^{\circ}$C]")
ax.set_ylabel(r"Humidity ratio ($m_{water}/m_{dry\ air}$) [-]")

# Saturation line
w = [HAProps('W', 'T', T, 'P', 101.325, 'R', 1.0) for T in Tdb]
ax.plot(Tdb - 273.15, w, lw=2)

# Humidity lines
RHValues = [0.05, 0.1, 0.15, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
for RH in RHValues:
    w = [HAProps('W', 'T', T, 'P', 101.325, 'R', RH) for T in Tdb]
    ax.plot(Tdb - 273.15, w, 'r', lw=1)

# Humidity lines
for H in [-20, -10, 0, 10, 20, 30, 40, 50, 60, 70, 80, 90]:
    # Line goes from saturation to zero humidity ratio for this enthalpy
    T1 = HAProps('T', 'H', H, 'P', 101.325, 'R', 1.0) - 273.15
    T0 = HAProps('T', 'H', H, 'P', 101.325, 'R', 0.0) - 273.15
    w1 = HAProps('W', 'H', H, 'P', 101.325, 'R', 1.0)
    w0 = HAProps('W', 'H', H, 'P', 101.325, 'R', 0.0)
    ax.plot(np.r_[T1, T0], np.r_[w1, w0], 'r', lw=1)

plt.show()



from __future__ import print_function


def run():
    import nose, os

    print('about to run the tests, please be patient')
    this_path, file = os.path.split(os.path.abspath(__file__))
    nose.run(argv=['--where', this_path])


if __name__ == '__main__':
    run()



# This file is automatically generated by the generate_constants_module.py script in wrappers/Python.
# DO NOT MODIFY THE CONTENTS OF THIS FILE!
from __future__ import absolute_import

from . import _constants

INVALID_PARAMETER = _constants.INVALID_PARAMETER
igas_constant = _constants.igas_constant
imolar_mass = _constants.imolar_mass
iacentric_factor = _constants.iacentric_factor
irhomolar_reducing = _constants.irhomolar_reducing
irhomolar_critical = _constants.irhomolar_critical
iT_reducing = _constants.iT_reducing
iT_critical = _constants.iT_critical
irhomass_reducing = _constants.irhomass_reducing
irhomass_critical = _constants.irhomass_critical
iP_critical = _constants.iP_critical
iP_reducing = _constants.iP_reducing
iT_triple = _constants.iT_triple
iP_triple = _constants.iP_triple
iT_min = _constants.iT_min
iT_max = _constants.iT_max
iP_max = _constants.iP_max
iP_min = _constants.iP_min
idipole_moment = _constants.idipole_moment
iT = _constants.iT
iP = _constants.iP
iQ = _constants.iQ
iTau = _constants.iTau
iDelta = _constants.iDelta
iDmolar = _constants.iDmolar
iHmolar = _constants.iHmolar
iSmolar = _constants.iSmolar
iCpmolar = _constants.iCpmolar
iCp0molar = _constants.iCp0molar
iCvmolar = _constants.iCvmolar
iUmolar = _constants.iUmolar
iGmolar = _constants.iGmolar
iHelmholtzmolar = _constants.iHelmholtzmolar
iSmolar_residual = _constants.iSmolar_residual
iDmass = _constants.iDmass
iHmass = _constants.iHmass
iSmass = _constants.iSmass
iCpmass = _constants.iCpmass
iCp0mass = _constants.iCp0mass
iCvmass = _constants.iCvmass
iUmass = _constants.iUmass
iGmass = _constants.iGmass
iHelmholtzmass = _constants.iHelmholtzmass
iviscosity = _constants.iviscosity
iconductivity = _constants.iconductivity
isurface_tension = _constants.isurface_tension
iPrandtl = _constants.iPrandtl
ispeed_sound = _constants.ispeed_sound
iisothermal_compressibility = _constants.iisothermal_compressibility
iisobaric_expansion_coefficient = _constants.iisobaric_expansion_coefficient
iisentropic_expansion_coefficient = _constants.iisentropic_expansion_coefficient
ifundamental_derivative_of_gas_dynamics = _constants.ifundamental_derivative_of_gas_dynamics
ialphar = _constants.ialphar
idalphar_dtau_constdelta = _constants.idalphar_dtau_constdelta
idalphar_ddelta_consttau = _constants.idalphar_ddelta_consttau
ialpha0 = _constants.ialpha0
idalpha0_dtau_constdelta = _constants.idalpha0_dtau_constdelta
idalpha0_ddelta_consttau = _constants.idalpha0_ddelta_consttau
iBvirial = _constants.iBvirial
iCvirial = _constants.iCvirial
idBvirial_dT = _constants.idBvirial_dT
idCvirial_dT = _constants.idCvirial_dT
iZ = _constants.iZ
iPIP = _constants.iPIP
ifraction_min = _constants.ifraction_min
ifraction_max = _constants.ifraction_max
iT_freeze = _constants.iT_freeze
iGWP20 = _constants.iGWP20
iGWP100 = _constants.iGWP100
iGWP500 = _constants.iGWP500
iFH = _constants.iFH
iHH = _constants.iHH
iPH = _constants.iPH
iODP = _constants.iODP
iPhase = _constants.iPhase
iundefined_parameter = _constants.iundefined_parameter
INPUT_PAIR_INVALID = _constants.INPUT_PAIR_INVALID
QT_INPUTS = _constants.QT_INPUTS
PQ_INPUTS = _constants.PQ_INPUTS
QSmolar_INPUTS = _constants.QSmolar_INPUTS
QSmass_INPUTS = _constants.QSmass_INPUTS
HmolarQ_INPUTS = _constants.HmolarQ_INPUTS
HmassQ_INPUTS = _constants.HmassQ_INPUTS
DmolarQ_INPUTS = _constants.DmolarQ_INPUTS
DmassQ_INPUTS = _constants.DmassQ_INPUTS
PT_INPUTS = _constants.PT_INPUTS
DmassT_INPUTS = _constants.DmassT_INPUTS
DmolarT_INPUTS = _constants.DmolarT_INPUTS
HmolarT_INPUTS = _constants.HmolarT_INPUTS
HmassT_INPUTS = _constants.HmassT_INPUTS
SmolarT_INPUTS = _constants.SmolarT_INPUTS
SmassT_INPUTS = _constants.SmassT_INPUTS
TUmolar_INPUTS = _constants.TUmolar_INPUTS
TUmass_INPUTS = _constants.TUmass_INPUTS
DmassP_INPUTS = _constants.DmassP_INPUTS
DmolarP_INPUTS = _constants.DmolarP_INPUTS
HmassP_INPUTS = _constants.HmassP_INPUTS
HmolarP_INPUTS = _constants.HmolarP_INPUTS
PSmass_INPUTS = _constants.PSmass_INPUTS
PSmolar_INPUTS = _constants.PSmolar_INPUTS
PUmass_INPUTS = _constants.PUmass_INPUTS
PUmolar_INPUTS = _constants.PUmolar_INPUTS
HmassSmass_INPUTS = _constants.HmassSmass_INPUTS
HmolarSmolar_INPUTS = _constants.HmolarSmolar_INPUTS
SmassUmass_INPUTS = _constants.SmassUmass_INPUTS
SmolarUmolar_INPUTS = _constants.SmolarUmolar_INPUTS
DmassHmass_INPUTS = _constants.DmassHmass_INPUTS
DmolarHmolar_INPUTS = _constants.DmolarHmolar_INPUTS
DmassSmass_INPUTS = _constants.DmassSmass_INPUTS
DmolarSmolar_INPUTS = _constants.DmolarSmolar_INPUTS
DmassUmass_INPUTS = _constants.DmassUmass_INPUTS
DmolarUmolar_INPUTS = _constants.DmolarUmolar_INPUTS
FLUID_TYPE_PURE = _constants.FLUID_TYPE_PURE
FLUID_TYPE_PSEUDOPURE = _constants.FLUID_TYPE_PSEUDOPURE
FLUID_TYPE_REFPROP = _constants.FLUID_TYPE_REFPROP
FLUID_TYPE_INCOMPRESSIBLE_LIQUID = _constants.FLUID_TYPE_INCOMPRESSIBLE_LIQUID
FLUID_TYPE_INCOMPRESSIBLE_SOLUTION = _constants.FLUID_TYPE_INCOMPRESSIBLE_SOLUTION
FLUID_TYPE_UNDEFINED = _constants.FLUID_TYPE_UNDEFINED
iphase_liquid = _constants.iphase_liquid
iphase_supercritical = _constants.iphase_supercritical
iphase_supercritical_gas = _constants.iphase_supercritical_gas
iphase_supercritical_liquid = _constants.iphase_supercritical_liquid
iphase_critical_point = _constants.iphase_critical_point
iphase_gas = _constants.iphase_gas
iphase_twophase = _constants.iphase_twophase
iphase_unknown = _constants.iphase_unknown
iphase_not_imposed = _constants.iphase_not_imposed
NORMALIZE_GAS_CONSTANTS = _constants.NORMALIZE_GAS_CONSTANTS
CRITICAL_WITHIN_1UK = _constants.CRITICAL_WITHIN_1UK
CRITICAL_SPLINES_ENABLED = _constants.CRITICAL_SPLINES_ENABLED
SAVE_RAW_TABLES = _constants.SAVE_RAW_TABLES
ALTERNATIVE_TABLES_DIRECTORY = _constants.ALTERNATIVE_TABLES_DIRECTORY
ALTERNATIVE_REFPROP_PATH = _constants.ALTERNATIVE_REFPROP_PATH
ALTERNATIVE_REFPROP_HMX_BNC_PATH = _constants.ALTERNATIVE_REFPROP_HMX_BNC_PATH
ALTERNATIVE_REFPROP_LIBRARY_PATH = _constants.ALTERNATIVE_REFPROP_LIBRARY_PATH
REFPROP_DONT_ESTIMATE_INTERACTION_PARAMETERS = _constants.REFPROP_DONT_ESTIMATE_INTERACTION_PARAMETERS
REFPROP_IGNORE_ERROR_ESTIMATED_INTERACTION_PARAMETERS = _constants.REFPROP_IGNORE_ERROR_ESTIMATED_INTERACTION_PARAMETERS
REFPROP_USE_GERG = _constants.REFPROP_USE_GERG
REFPROP_USE_PENGROBINSON = _constants.REFPROP_USE_PENGROBINSON
MAXIMUM_TABLE_DIRECTORY_SIZE_IN_GB = _constants.MAXIMUM_TABLE_DIRECTORY_SIZE_IN_GB
DONT_CHECK_PROPERTY_LIMITS = _constants.DONT_CHECK_PROPERTY_LIMITS
HENRYS_LAW_TO_GENERATE_VLE_GUESSES = _constants.HENRYS_LAW_TO_GENERATE_VLE_GUESSES
PHASE_ENVELOPE_STARTING_PRESSURE_PA = _constants.PHASE_ENVELOPE_STARTING_PRESSURE_PA
R_U_CODATA = _constants.R_U_CODATA
VTPR_UNIFAC_PATH = _constants.VTPR_UNIFAC_PATH
SPINODAL_MINIMUM_DELTA = _constants.SPINODAL_MINIMUM_DELTA
OVERWRITE_FLUIDS = _constants.OVERWRITE_FLUIDS
OVERWRITE_DEPARTURE_FUNCTION = _constants.OVERWRITE_DEPARTURE_FUNCTION
OVERWRITE_BINARY_INTERACTION = _constants.OVERWRITE_BINARY_INTERACTION
USE_GUESSES_IN_PROPSSI = _constants.USE_GUESSES_IN_PROPSSI
ASSUME_CRITICAL_POINT_STABLE = _constants.ASSUME_CRITICAL_POINT_STABLE
VTPR_ALWAYS_RELOAD_LIBRARY = _constants.VTPR_ALWAYS_RELOAD_LIBRARY
FLOAT_PUNCTUATION = _constants.FLOAT_PUNCTUATION



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import matplotlib.pyplot as plt
import numpy as np
from abc import ABCMeta
from six import with_metaclass
import warnings

import CoolProp
from CoolProp import AbstractState
from CoolProp import CoolProp as CP
from CoolProp.CoolProp import PropsSI, extract_backend, extract_fractions, PyCriticalState


def get_critical_point(state):
    crit_state = PyCriticalState()
    crit_state.T = np.nan
    crit_state.p = np.nan
    crit_state.rhomolar = np.nan
    crit_state.rhomolar = np.nan
    crit_state.stable = False
    try:
        crit_state.T = state.T_critical()
        crit_state.p = state.p_critical()
        crit_state.rhomolar = state.rhomolar_critical()
        crit_state.stable = True
    except:
        try:
            for crit_state_tmp in state.all_critical_points():
                if crit_state_tmp.stable and (crit_state_tmp.T > crit_state.T or not np.isfinite(crit_state.T)):
                    crit_state.T = crit_state_tmp.T
                    crit_state.p = crit_state_tmp.p
                    crit_state.rhomolar = crit_state_tmp.rhomolar
                    crit_state.stable = crit_state_tmp.stable
        except:
            raise ValueError("Could not calculate the critical point data.")
    new_state = AbstractState(state.backend_name(), '&'.join(state.fluid_names()))
    masses = state.get_mass_fractions()
    if len(masses) > 1:
        new_state.set_mass_fractions(masses)  # Uses mass fraction to work with incompressibles
        # try: new_state.build_phase_envelope("dummy")
        # except: pass
    msg = ""
    if np.isfinite(crit_state.p) and np.isfinite(crit_state.T):
        try:
            new_state.specify_phase(CoolProp.iphase_critical_point)
            new_state.update(CoolProp.PT_INPUTS, crit_state.p, crit_state.T)
            return new_state
        except Exception as e:
            msg += str(e) + " - "
            pass
        try:
            new_state.update(CoolProp.PT_INPUTS, crit_state.p, crit_state.T)
            return new_state
        except Exception as e:
            msg += str(e) + " - "
            pass
    if np.isfinite(crit_state.rhomolar) and np.isfinite(crit_state.T):
        try:
            new_state.specify_phase(CoolProp.iphase_critical_point)
            new_state.update(CoolProp.DmolarT_INPUTS, crit_state.rhomolar, crit_state.T)
            return new_state
        except Exception as e:
            msg += str(e) + " - "
            pass
        try:
            new_state.update(CoolProp.DmolarT_INPUTS, crit_state.rhomolar, crit_state.T)
            return new_state
        except Exception as e:
            msg += str(e) + " - "
            pass
    raise ValueError("Could not calculate the critical point data. " + msg)


def interpolate_values_1d(x, y, x_points=None, kind='linear'):
    try:
        from scipy.interpolate.interpolate import interp1d
        if x_points is None:
            return interp1d(x, y, kind=kind)(x[np.isfinite(x)])
        else:
            return interp1d(x, y, kind=kind)(x_points)
    except ImportError:
        if kind != 'linear':
            warnings.warn(
              "You requested a non-linear interpolation, but SciPy is not available. Falling back to linear interpolation.",
              UserWarning)
        if x_points is None:
            return np.interp((x[np.isfinite(x)]), x, y)
        else:
            return np.interp(x_points, x, y)


def is_string(in_obj):
    try:
        return isinstance(in_obj, basestring)
    except NameError:
        return isinstance(in_obj, str)
    # except:
    #    return False


def process_fluid_state(fluid_ref, fractions='mole'):
    """Check input for state object or fluid string

    Parameters
    ----------
        fluid_ref : str, CoolProp.AbstractState
        fractions : str, switch to set mass, volu or mole fractions

    Returns
    -------
        CoolProp.AbstractState
    """
    # Process the fluid and set self._state
    if is_string(fluid_ref):
        backend, fluids = extract_backend(fluid_ref)
        fluids, fractions = extract_fractions(fluids)
        state = AbstractState(backend, '&'.join(fluids))
        if len(fluids) > 1 and len(fluids) == len(fractions):
            if fractions == 'mass': state.set_mass_fractions(fractions)
            elif fractions == 'volu': state.set_volu_fractions(fractions)
            else: state.set_mole_fractions(fractions)
        return state
    elif isinstance(fluid_ref, AbstractState):
        return fluid_ref
    raise TypeError("Invalid fluid_ref input, expected a string or an abstract state instance.")


def _get_index(prop):
    if is_string(prop):
        return CP.get_parameter_index(prop)
    elif isinstance(prop, int):
        return prop
    else:
        raise ValueError("Invalid input, expected a string or an int, not {0:s}.".format(str(prop)))


class BaseQuantity(object):
    """A very basic property that can convert an input to and from a 
    given unit system, note that the conversion from SI units starts 
    with a multiplication. If you need to remove an offset, use the 
    off_SI property. 
    Examples with temperature:
    celsius = BaseQuantity(add_SI=-273.15)
    fahrenheit = BaseQuantity(add_SI=32.0, mul_SI=1.8, off_SI=-273.15)
    Examples with pressure:
    bar = BaseQuantity(mul_SI=1e-5)
    psi = BaseQuantity(mul_SI=0.000145037738)    
    """

    def __init__(self, add_SI=0.0, mul_SI=1.0, off_SI=0.0):
        self._add_SI = add_SI
        self._mul_SI = mul_SI
        self._off_SI = off_SI

    @property
    def add_SI(self): return self._add_SI

    @add_SI.setter
    def add_SI(self, value): self._add_SI = value

    @property
    def mul_SI(self): return self._mul_SI

    @mul_SI.setter
    def mul_SI(self, value): self._mul_SI = value

    @property
    def off_SI(self): return self._off_SI

    @off_SI.setter
    def off_SI(self, value): self._off_SI = value

    def from_SI(self, value): return ((value + self.off_SI) * self.mul_SI) + self.add_SI

    def to_SI(self, value): return (value - self.add_SI) / self.mul_SI - self.off_SI


class BaseDimension(BaseQuantity):
    """A dimension is a class that extends the BaseQuantity and adds a label, a symbol and a unit label"""

    def __init__(self, add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='', symbol='', unit=''):
        self._label = label
        self._symbol = symbol
        self._unit = unit
        super(BaseDimension, self).__init__(add_SI=add_SI, mul_SI=mul_SI, off_SI=off_SI)

    @property
    def label(self): return self._label

    @label.setter
    def label(self, value): self._label = value

    @property
    def symbol(self): return self._symbol

    @symbol.setter
    def symbol(self, value): self._symbol = value

    @property
    def unit(self): return self._unit

    @unit.setter
    def unit(self, value): self._unit = value


class PropertyDict(with_metaclass(ABCMeta), object):
    """A collection of dimensions for all the required quantities"""

    def __init__(self):
        self._D = None
        self._H = None
        self._P = None
        self._S = None
        self._T = None
        self._U = None
        self._Q = None

    @property
    def D(self): return self._D

    @D.setter
    def D(self, value): self._D = value

    @property
    def H(self): return self._H

    @H.setter
    def H(self, value): self._H = value

    @property
    def P(self): return self._P

    @P.setter
    def P(self, value): self._P = value

    @property
    def S(self): return self._S

    @S.setter
    def S(self, value): self._S = value

    @property
    def T(self): return self._T

    @T.setter
    def T(self, value): self._T = value

    @property
    def U(self): return self._U

    @U.setter
    def U(self, value): self._U = value

    @property
    def Q(self): return self._Q

    @Q.setter
    def Q(self, value): self._Q = value

    @property
    def dimensions(self):
        return {
      CoolProp.iDmass: self._D,
      CoolProp.iHmass: self._H,
      CoolProp.iP: self._P,
      CoolProp.iSmass: self._S,
      CoolProp.iT: self._T,
      CoolProp.iUmass: self._U,
      CoolProp.iQ: self._Q
    }

    def __getitem__(self, index):
        """Allow for property access via square brackets"""
        idx = _get_index(index)
        if idx == CoolProp.iDmass: return self.D
        elif idx == CoolProp.iHmass: return self.H
        elif idx == CoolProp.iP: return self.P
        elif idx == CoolProp.iSmass: return self.S
        elif idx == CoolProp.iT: return self.T
        elif idx == CoolProp.iUmass: return self.U
        elif idx == CoolProp.iQ: return self.Q
        else: raise IndexError("Unknown index \"{0:s}\".".format(str(index)))

    def __setitem__(self, index, value):
        """Allow for property access via square brackets"""
        idx = _get_index(index)
        if idx == CoolProp.iDmass: self.D = value
        elif idx == CoolProp.iHmass: self.H = value
        elif idx == CoolProp.iP: self.P = value
        elif idx == CoolProp.iSmass: self.S = value
        elif idx == CoolProp.iT: self.T = value
        elif idx == CoolProp.iUmass: self.U = value
        elif idx == CoolProp.iQ: self.Q = value
        else: raise IndexError("Unknown index \"{0:s}\".".format(str(index)))


class SIunits(PropertyDict):
    def __init__(self):
        self._D = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Density', symbol=u'd', unit=u'kg/m3')
        self._H = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Specific Enthalpy', symbol=u'h', unit=u'J/kg')
        self._P = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Pressure', symbol=u'p', unit=u'Pa')
        self._S = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Specific Entropy', symbol=u's', unit=u'J/kg/K')
        self._T = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Temperature', symbol=u'T', unit=u'K')
        self._U = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Specific Internal Energy', symbol=u'u', unit=u'J/kg')
        self._Q = BaseDimension(add_SI=0.0, mul_SI=1.0, off_SI=0.0, label='Vapour Quality', symbol=u'x', unit=u'')


class KSIunits(SIunits):
    def __init__(self):
        super(KSIunits, self).__init__()
        self.H.mul_SI = 1e-3
        self.H.unit = u'kJ/kg'
        self.P.mul_SI = 1e-3
        self.P.unit = u'kPa'
        self.S.mul_SI = 1e-3
        self.S.unit = u'kJ/kg/K'
        self.U.mul_SI = 1e-3
        self.U.unit = u'kJ/kg'


class EURunits(KSIunits):
    def __init__(self):
        super(EURunits, self).__init__()
        self.P.mul_SI = 1e-5
        self.P.unit = u'bar'
        self.T.add_SI = -273.15
        self.T.unit = u'deg C'


class Base2DObject(with_metaclass(ABCMeta), object):
    """A container for shared settings and constants for the 
    isolines and the property plots."""

    # A list of supported plot
    TS = CoolProp.iT * 10 + CoolProp.iSmass
    PH = CoolProp.iP * 10 + CoolProp.iHmass
    HS = CoolProp.iHmass * 10 + CoolProp.iSmass
    PS = CoolProp.iP * 10 + CoolProp.iSmass
    PD = CoolProp.iP * 10 + CoolProp.iDmass
    TD = CoolProp.iT * 10 + CoolProp.iDmass
    PT = CoolProp.iP * 10 + CoolProp.iT
    PU = CoolProp.iP * 10 + CoolProp.iUmass

    PLOTS = {
      'TS': TS,
      'PH': PH,
      'HS': HS,
      'PS': PS,
      'PD': PD,
      'TD': TD,
      'PT': PT,
    }

    PLOTS_INV = {v: k for k, v in PLOTS.items()}

#     # A list of supported plot
#     @property
#     def TS(self): return type(self).TS
#     @property
#     def PH(self): return CoolProp.iP*10     + CoolProp.iHmass
#     @property
#     def HS(self): return CoolProp.iHmass*10 + CoolProp.iSmass
#     @property
#     def PS(self): return CoolProp.iP*10     + CoolProp.iSmass
#     @property
#     def PD(self): return CoolProp.iP*10     + CoolProp.iDmass
#     @property
#     def TD(self): return CoolProp.iT*10     + CoolProp.iDmass
#     @property
#     def PT(self): return CoolProp.iP*10     + CoolProp.iT
#     @property
#     def PU(self): return CoolProp.iP*10     + CoolProp.iUmass

    def __init__(self, x_type, y_type, state=None, small=None):
        self._x_index = _get_index(x_type)
        self._y_index = _get_index(y_type)
        self._critical_state = None
        if small is not None: self._small = small
        else: self._small = 1e-7
        if state is not None: self.state = state
        else: self._state = None

    # A list of supported plot
    @property
    def x_index(self): return self._x_index

    @property
    def y_index(self): return self._y_index

    @property
    def critical_state(self):
        if self._critical_state is None and self._state is not None:
            self._critical_state = get_critical_point(self._state)
        return self._critical_state

    @property
    def state(self): return self._state

    @state.setter
    def state(self, value):
        self._state = process_fluid_state(value)
        # try: self._state.build_phase_envelope("dummy")
        # except: pass
        self._critical_state = None
        #self._T_small = self._state.trivial_keyed_output(CoolProp.iT_critical)*self._small
        #self._P_small = self._state.trivial_keyed_output(CoolProp.iP_critical)*self._small
        self._T_small = self.critical_state.keyed_output(CoolProp.iT) * self._small
        self._P_small = self.critical_state.keyed_output(CoolProp.iP) * self._small

    def _get_sat_bounds(self, kind, smin=None, smax=None):
        """Generates limits for the saturation line in either T or p determined
        by 'kind'. If smin or smax are provided, values will be checked
        against the allowable range for the EOS and a warning might be
        generated. Returns a tuple containing (xmin, xmax)"""

        # TODO: REFPROP backend does not have ptriple.
        T_triple = self._state.trivial_keyed_output(CoolProp.iT_triple)
        try:
            T_min = self._state.trivial_keyed_output(CoolProp.iT_min)
        except:
            T_min = T_triple
        self._state.update(CoolProp.QT_INPUTS, 0, max([T_triple, T_min]) + self._T_small)
        kind = _get_index(kind)
        if kind == CoolProp.iP:
            fluid_min = self._state.keyed_output(CoolProp.iP) + self._P_small
            fluid_max = self.critical_state.keyed_output(CoolProp.iP) - self._P_small
        elif kind == CoolProp.iT:
            fluid_min = self._state.keyed_output(CoolProp.iT) + self._T_small
            fluid_max = self.critical_state.keyed_output(CoolProp.iT) - self._T_small
        else:
            raise ValueError("Saturation boundaries have to be defined in T or P, but not in {0:s}".format(str(kind)))

        if smin is not None:
            if fluid_min < smin < fluid_max:
                sat_min = smin
            else:
                warnings.warn(
                  "Your minimum {0:s} has been ignored, {1:f} is not between {2:f} and {3:f}".format(self.PROPERTIES[kind], smin, fluid_min, fluid_max),
                  UserWarning)
                sat_min = fluid_min
        else:
            sat_min = fluid_min

        if smax is not None:
            if fluid_min < smax < fluid_max:
                sat_max = smax
            else:
                warnings.warn(
                  "Your maximum {0:s} has been ignored, {1:f} is not between {2:f} and {3:f}".format(self.PROPERTIES[kind], smax, fluid_min, fluid_max),
                  UserWarning)
                sat_max = fluid_max
        else:
            sat_max = fluid_max

        return sat_min, sat_max


class IsoLine(Base2DObject):
    """An object that holds the functions to calculate a line of 
    a constant property in the dimensions of a property plot. This 
    class only uses SI units."""

    # Normally we calculate a sweep in x-dimensions, but
    # sometimes a sweep in y-dimensions is better.
    XY_SWITCH = {
      CoolProp.iDmass: {Base2DObject.TS: True, Base2DObject.PH: True, Base2DObject.HS: False, Base2DObject.PS: True, Base2DObject.PD: None, Base2DObject.TD: None, Base2DObject.PT: False},
      CoolProp.iHmass: {Base2DObject.TS: False, Base2DObject.PH: None, Base2DObject.HS: None, Base2DObject.PS: True, Base2DObject.PD: True, Base2DObject.TD: False, Base2DObject.PT: False},
      CoolProp.iP: {Base2DObject.TS: False, Base2DObject.PH: None, Base2DObject.HS: False, Base2DObject.PS: None, Base2DObject.PD: None, Base2DObject.TD: False, Base2DObject.PT: None},
      CoolProp.iSmass: {Base2DObject.TS: None, Base2DObject.PH: True, Base2DObject.HS: None, Base2DObject.PS: None, Base2DObject.PD: True, Base2DObject.TD: False, Base2DObject.PT: True},
      CoolProp.iT: {Base2DObject.TS: None, Base2DObject.PH: True, Base2DObject.HS: False, Base2DObject.PS: False, Base2DObject.PD: False, Base2DObject.TD: None, Base2DObject.PT: None},
      CoolProp.iQ: {Base2DObject.TS: True, Base2DObject.PH: True, Base2DObject.HS: True, Base2DObject.PS: True, Base2DObject.PD: True, Base2DObject.TD: True, Base2DObject.PT: False}
    }

    # Abort interpolation if there are not enough
    # valid entries.
    VALID_REQ = 5.0 / 100.0

    def __init__(self, i_index, x_index, y_index, value=0.0, state=None):
        super(IsoLine, self).__init__(x_index, y_index, state)
        self._i_index = _get_index(i_index)
        if value is not None: self.value = value
        else: self._value = None
        self._x = None
        self._y = None

    @property
    def i_index(self): return self._i_index

    @property
    def value(self): return self._value

    @value.setter
    def value(self, value): self._value = float(value)

    @property
    def x(self): return self._x

    @x.setter
    def x(self, value): self._x = np.array(value)

    @property
    def y(self): return self._y

    @y.setter
    def y(self, value): self._y = np.array(value)

    def get_update_pair(self):
        """Processes the values for the isoproperty and the graph dimensions
        to figure which should be used as inputs to the state update. Returns
        a tuple with the indices for the update call and the property constant.
        For an isobar in a Ts-diagram it returns the default order and the 
        correct constant for the update pair:
        get_update_pair(CoolProp.iP,CoolProp.iSmass,CoolProp.iT) -> (0,1,2,CoolProp.PSmass_INPUTS)
        other values require switching and swapping.
        """
        # Figure out if x or y-dimension should be used
        switch = self.XY_SWITCH[self.i_index][self.y_index * 10 + self.x_index]

        if switch is None:
            raise ValueError("This isoline cannot be calculated!")
        elif switch is False:
            pair, out1, _ = CP.generate_update_pair(self.i_index, 0.0, self.x_index, 1.0)
        elif switch is True:
            pair, out1, _ = CP.generate_update_pair(self.i_index, 0.0, self.y_index, 1.0)
        else:
            raise ValueError("Unknown error!")

        if out1 == 0.0:  # Correct order
            swap = False
        else:  # Wrong order
            swap = True

        if not switch and not swap:
            return 0, 1, 2, pair
        elif switch and not swap:
            return 0, 2, 1, pair
        elif not switch and swap:
            return 1, 0, 2, pair
        elif switch and swap:
            return 1, 2, 0, pair
        else:
            raise ValueError("Check the code, this should not happen!")

    def calc_sat_range(self, Trange=None, Prange=None, num=200):
        if Trange is not None:
            two = np.array(Trange)
            one = np.resize(np.array(self.value), two.shape)
            pair = CoolProp.QT_INPUTS
        elif Prange is not None:
            one = np.array(Prange)
            two = np.resize(np.array(self.value), one.shape)
            pair = CoolProp.PQ_INPUTS
        else:
            T_lo, T_hi = self._get_sat_bounds(CoolProp.iT)
            two = np.linspace(T_lo, T_hi, num)
            one = np.resize(np.array(self.value), two.shape)
            pair = CoolProp.QT_INPUTS

        Tcrit = self.critical_state.keyed_output(CoolProp.iT)
        Pcrit = self.critical_state.keyed_output(CoolProp.iP)
        Dcrit = self.critical_state.keyed_output(CoolProp.iDmass)
        try:
            #self.state.update(CoolProp.DmassT_INPUTS, Dcrit, Tcrit)
            #xcrit = self.state.keyed_output(self._x_index)
            #ycrit = self.state.keyed_output(self._y_index)
            xcrit = self.critical_state.keyed_output(self._x_index)
            ycrit = self.critical_state.keyed_output(self._y_index)
        except:
            warnings.warn(
              "An error occurred for the critical inputs, skipping it.",
              UserWarning)
            xcrit = np.NaN
            ycrit = np.NaN

        X = np.empty_like(one)
        Y = np.empty_like(one)

        err = False
        for index, _ in np.ndenumerate(one):
            try:
                self.state.update(pair, one[index], two[index])
                X[index] = self.state.keyed_output(self._x_index)
                Y[index] = self.state.keyed_output(self._y_index)
            except Exception as e:
                if (pair == CoolProp.QT_INPUTS and abs(two[index] - Tcrit) < 1e0) or \
                   (pair == CoolProp.PQ_INPUTS and abs(one[index] - Pcrit) < 1e2):
                    X[index] = xcrit
                    Y[index] = ycrit
                    warnings.warn(
                  "An error occurred for near critical inputs {0:f}, {1:f} with index {2:s}: {3:s}".format(one[index], two[index], str(index), str(e)),
                  UserWarning)
                    pass

                warnings.warn(
                  "An error occurred for inputs {0:f}, {1:f} with index {2:s}: {3:s}".format(one[index], two[index], str(index), str(e)),
                  UserWarning)
                X[index] = np.NaN
                Y[index] = np.NaN
                err = True
        self.x = X; self.y = Y
        return

    def calc_range(self, xvals=None, yvals=None):

        if self.i_index == CoolProp.iQ:
            warnings.warn(
                "Please use \"calc_sat_range\" to calculate saturation and isoquality lines. Input ranges are discarded.",
                UserWarning)
            if xvals is not None: self.calc_sat_range(num=xvals.size)
            elif yvals is not None: self.calc_sat_range(num=yvals.size)
            else: self.calc_sat_range()
            return

        ipos, xpos, ypos, pair = self.get_update_pair()

        order = [ipos, xpos, ypos]
        idxs = [v for (_, v) in sorted(zip(order, [self.i_index, self.x_index, self.y_index]))]
        vals = [v for (_, v) in sorted(zip(order, [np.array(self.value), xvals, yvals]))]
        if vals[0] is None or vals[1] is None:
            raise ValueError("One required input is missing, make sure to supply the correct xvals ({0:s}) or yvals ({1:s}).".format(str(xvals), str(yvals)))

        if vals[0].size > vals[1].size:
            vals[1] = np.resize(vals[1], vals[0].shape)
        elif vals[0].size < vals[1].size:
            vals[0] = np.resize(vals[0], vals[1].shape)

        vals[2] = np.empty_like(vals[0])
        err = False
        guesses = CoolProp.CoolProp.PyGuessesStructure()
        # Only use the guesses for selected inputs
        if pair == CoolProp.HmolarP_INPUTS \
          or pair == CoolProp.HmassP_INPUTS:
            # or pair == CoolProp.HmassSmass_INPUTS \
            # or pair == CoolProp.HmolarSmolar_INPUTS \
            # or pair == CoolProp.PSmass_INPUTS \
            # or pair == CoolProp.PSmolar_INPUTS:
            use_guesses = True
        else:
            use_guesses = False
        for index, _ in np.ndenumerate(vals[0]):
            try:
                if use_guesses:
                    if np.isfinite(guesses.rhomolar):
                        self.state.update_with_guesses(pair, vals[0][index], vals[1][index], guesses)
                    else:
                        self.state.update(pair, vals[0][index], vals[1][index])
                    guesses.rhomolar = self.state.rhomolar()
                    guesses.T = self.state.T()
                else:
                    self.state.update(pair, vals[0][index], vals[1][index])
                vals[2][index] = self.state.keyed_output(idxs[2])
            except Exception as e:
                warnings.warn(
                  "An error occurred for inputs {0:f}, {1:f} with index {2:s}: {3:s}".format(vals[0][index], vals[1][index], str(index), str(e)),
                  UserWarning)
                vals[2][index] = np.NaN
                guesses.rhomolar = np.NaN
                guesses.T = np.NaN
                err = True

        for i, v in enumerate(idxs):
            if v == self.x_index: self.x = vals[i]
            if v == self.y_index: self.y = vals[i]

    def sanitize_data(self):
        """Fill the series via interpolation"""
        validx = None; validy = None
        countx = None; county = None
        if self.x is not None:
            validx = np.isfinite(self.x)
            countx = float(self.x.size)
        else:
            raise ValueError("The x-axis is not populated, calculate values before you interpolate.")
        if self.y is not None:
            validy = np.isfinite(self.y)
            county = float(self.y.size)
        else:
            raise ValueError("The y-axis is not populated, calculate values before you interpolate.")

        if min([np.sum(validx) / countx, np.sum(validy) / county]) < self.VALID_REQ:
            warnings.warn(
              "Poor data quality, there are not enough valid entries for x ({0:f}/{1:f}) or y ({2:f}/{3:f}).".format(np.sum(validx), countx, np.sum(validy), county),
              UserWarning)
        # TODO: use filter and cubic splines!
        #filter = np.logical_and(np.isfinite(self.x),np.isfinite(self.y))
        if np.sum(validy) > np.sum(validx):
            self.x = interpolate_values_1d(self.y, self.x, x_points=self.y[validy])
            self.y = self.y[validy]
        else:
            self.y = interpolate_values_1d(self.x, self.y, x_points=self.x[validx])
            self.x = self.x[validx]


class BasePlot(Base2DObject):
    """The base class for all plots. It can be instantiated itself, but provides many 
    general facilities to be used in the different plots. """

    # Define the iteration keys
    PROPERTIES = {
      CoolProp.iDmass: 'density',
      CoolProp.iHmass: 'specific enthalpy',
      CoolProp.iP: 'pressure',
      CoolProp.iSmass: 'specific entropy',
      CoolProp.iT: 'temperature',
      CoolProp.iUmass: 'specific internal energy'
    }

    # Define the unit systems
    UNIT_SYSTEMS = {
      'SI': SIunits(),
      'KSI': KSIunits(),
      'EUR': EURunits()
    }

    LINE_PROPS = {
      CoolProp.iT: dict(color='Darkred', lw=0.25),
      CoolProp.iP: dict(color='DarkCyan', lw=0.25),
      CoolProp.iHmass: dict(color='DarkGreen', lw=0.25),
      CoolProp.iDmass: dict(color='DarkBlue', lw=0.25),
      CoolProp.iSmass: dict(color='DarkOrange', lw=0.25),
      CoolProp.iQ: dict(color='black', lw=0.25)
    }

    ID_FACTOR = 10.0  # Values below this number are interpreted as factors
    HI_FACTOR = 2.25  # Upper default limits: HI_FACTOR*T_crit and HI_FACTOR*p_crit
    LO_FACTOR = 1.01  # Lower default limits: LO_FACTOR*T_triple and LO_FACTOR*p_triple

    TP_LIMITS = {
      'NONE': [None, None, None, None],
      'DEF': [LO_FACTOR, HI_FACTOR, LO_FACTOR, HI_FACTOR],
      'ACHP': [173.15, 493.15, 0.25e5, HI_FACTOR],
      'ORC': [273.15, 673.15, 0.25e5, HI_FACTOR]
    }

    def __init__(self, fluid_ref, graph_type, unit_system='KSI', tp_limits='DEF', **kwargs):

        # Process the graph_type and set self._x_type and self._y_type
        graph_type = graph_type.upper()
        graph_type = graph_type.replace(r'RHO', r'D')
        if graph_type not in Base2DObject.PLOTS:
            raise ValueError("Invalid graph_type input, expected a string from {0:s}".format(str(self.PLOTS)))

        # Process the unit_system and set self._system
        self.system = unit_system
        # Process the plotting range based on T and p
        self.limits = tp_limits
        # Other properties
        self.figure = kwargs.pop('figure', plt.figure(tight_layout=True))
        self.axis = kwargs.pop('axis', self.figure.add_subplot(111))
        self.props = kwargs.pop('props', None)

        # call the base class
        state = process_fluid_state(fluid_ref)
        Base2DObject.__init__(self, graph_type[1], graph_type[0], state, **kwargs)

    @property
    def system(self): return self._system

    @system.setter
    def system(self, value):
        value = value.upper()
        if value in self.UNIT_SYSTEMS: self._system = self.UNIT_SYSTEMS[value]
        else: raise ValueError("Invalid input, expected a string from {0:s}".format(str(self.UNIT_SYSTEMS.keys())))

    @property
    def limits(self):
        """Returns [Tmin,Tmax,pmin,pmax] as value or factors"""
        return self._limits

    @limits.setter
    def limits(self, value):
        if is_string(value):
            value = value.upper()
        if value in self.TP_LIMITS:
            self._limits = self.TP_LIMITS[value]
        elif len(value) == 4:
            self._limits = value
        else:
            raise ValueError("Invalid input, expected a list with 4 items or a string from {0:s}".format(str(self.TP_LIMITS.keys())))

    @property
    def figure(self): return self._figure

    @figure.setter
    def figure(self, value): self._figure = value

    @property
    def axis(self): return self._axis

    @axis.setter
    def axis(self, value): self._axis = value

    @property
    def props(self): return self._props

    @props.setter
    def props(self, value):
        self._props = self.LINE_PROPS.copy()
        if value is not None:
            self._props.update(value)

    def __sat_bounds(self, kind, smin=None, smax=None):
        warnings.warn(
          "You called the deprecated function \"__sat_bounds\", \
consider replacing it with \"_get_sat_bounds\".",
          DeprecationWarning)
        return self._get_sat_bounds(kind, smin, smax)

    def _get_iso_label(self, isoline, unit=True):
        if self._system is not None:
            dim = self._system[isoline.i_index]
            return str(r"$" + dim.symbol + "=" + str(dim.from_SI(isoline.value)) + "$ " + dim.unit if unit else "$").strip()
        return str(isoline.value).strip()

    # def _get_phase_envelope(self):
    #
    #HEOS = CoolProp.AbstractState("HEOS", fluid)
    # HEOS.build_phase_envelope("")
    #PED = HEOS.get_phase_envelope_data()
    #plt.plot(PED.T, np.log(PED.p))
    # plt.show()

    def _plot_default_annotations(self):
#         def filter_fluid_ref(fluid_ref):
#             fluid_ref_string = fluid_ref
#             if fluid_ref.startswith('REFPROP-MIX'):
#                 end = 0
#                 fluid_ref_string = ''
#                 while fluid_ref.find('[', end + 1) != -1:
#                     start = fluid_ref.find('&', end + 1)
#                     if end == 0:
#                         start = fluid_ref.find(':', end + 1)
#                     end = fluid_ref.find('[', end + 1)
#                     fluid_ref_string = ' '.join([fluid_ref_string,
#                                                 fluid_ref[start+1:end], '+'])
#                 fluid_ref_string = fluid_ref_string[0:len(fluid_ref_string)-2]
#             return fluid_ref_string
#
#         if len(self.graph_type) == 2:
#             y_axis_id = self.graph_type[0]
#             x_axis_id = self.graph_type[1]
#         else:
#             y_axis_id = self.graph_type[0]
#             x_axis_id = self.graph_type[1:len(self.graph_type)]
#
#         tl_str = "%s - %s Graph for %s"
#         if not self.axis.get_title():
#             self.axis.set_title(tl_str % (self.AXIS_LABELS[self.unit_system][y_axis_id][0],
#                                           self.AXIS_LABELS[self.unit_system][x_axis_id][0],
#                                           filter_fluid_ref(self.fluid_ref)))
        if self._x_index in [CoolProp.iDmass, CoolProp.iP]:
            self.axis.set_xscale('log')
        if self._y_index in [CoolProp.iDmass, CoolProp.iP]:
            self.axis.set_yscale('log')

        if not self.axis.get_xlabel():
            dim = self._system[self._x_index]
            self.xlabel((dim.label + u" $" + dim.symbol + u"$ / " + dim.unit).strip())
        if not self.axis.get_ylabel():
            dim = self._system[self._y_index]
            self.ylabel((dim.label + u" $" + dim.symbol + u"$ / " + dim.unit).strip())

    def title(self, title):
        self.axis.set_title(title)

    def xlabel(self, xlabel):
        self.axis.set_xlabel(xlabel)

    def ylabel(self, ylabel):
        self.axis.set_ylabel(ylabel)

    def grid(self, b=None, **kwargs):
        g_map = {'on': True, 'off': False}
        if b is not None:
            b = g_map[b.lower()]
        if not kwargs:  # len=0
            self.axis.grid(b)
        else:
            self.axis.grid(kwargs)

    def set_Tp_limits(self, limits):
        """Set the limits for the graphs in temperature and pressure, based on 
        the active units: [Tmin, Tmax, pmin, pmax]"""
        dim = self._system[CoolProp.iT]
        limits[0] = dim.to_SI(limits[0])
        limits[1] = dim.to_SI(limits[1])
        dim = self._system[CoolProp.iP]
        limits[2] = dim.to_SI(limits[2])
        limits[3] = dim.to_SI(limits[3])
        self.limits = limits

    def get_Tp_limits(self):
        """Get the limits for the graphs in temperature and pressure, based on 
        the active units: [Tmin, Tmax, pmin, pmax]"""
        limits = self._get_Tp_limits()
        dim = self._system[CoolProp.iT]
        limits[0] = dim.from_SI(limits[0])
        limits[1] = dim.from_SI(limits[1])
        dim = self._system[CoolProp.iP]
        limits[2] = dim.from_SI(limits[2])
        limits[3] = dim.from_SI(limits[3])
        return limits

    def _get_Tp_limits(self):
        """Get the limits for the graphs in temperature and pressure, based on 
        SI units: [Tmin, Tmax, pmin, pmax]"""
        T_lo, T_hi, P_lo, P_hi = self.limits
        Ts_lo, Ts_hi = self._get_sat_bounds(CoolProp.iT)
        Ps_lo, Ps_hi = self._get_sat_bounds(CoolProp.iP)

        if T_lo is None: T_lo = 0.0
        elif T_lo < self.ID_FACTOR: T_lo *= Ts_lo
        if T_hi is None: T_hi = 1e6
        elif T_hi < self.ID_FACTOR: T_hi *= Ts_hi
        if P_lo is None: P_lo = 0.0
        elif P_lo < self.ID_FACTOR: P_lo *= Ps_lo
        if P_hi is None: P_hi = 1e10
        elif P_hi < self.ID_FACTOR: P_hi *= Ps_hi

        try: T_lo = np.nanmax([T_lo, self._state.trivial_keyed_output(CoolProp.iT_min)])
        except: pass
        try: T_hi = np.nanmin([T_hi, self._state.trivial_keyed_output(CoolProp.iT_max)])
        except: pass
        try: P_lo = np.nanmax([P_lo, self._state.trivial_keyed_output(CoolProp.iP_min)])
        except: pass
        try: P_hi = np.nanmin([P_hi, self._state.trivial_keyed_output(CoolProp.iP_max)])
        except: pass

        return [T_lo, T_hi, P_lo, P_hi]

    def set_axis_limits(self, limits):
        """Set the limits of the internal axis object based on the active units,
        takes [xmin, xmax, ymin, ymax]"""
        self.axis.set_xlim([limits[0], limits[1]])
        self.axis.set_ylim([limits[2], limits[3]])

    def _set_axis_limits(self, limits):
        """Set the limits of the internal axis object based on SI units,
        takes [xmin, xmax, ymin, ymax]"""
        dim = self._system[self._x_index]
        self.axis.set_xlim([dim.from_SI(limits[0]), dim.from_SI(limits[1])])
        dim = self._system[self._y_index]
        self.axis.set_ylim([dim.from_SI(limits[2]), dim.from_SI(limits[3])])

    def get_axis_limits(self, x_index=None, y_index=None):
        """Returns the previously set limits or generates them and 
        converts the default values to the selected unit system.
        Returns a list containing [xmin, xmax, ymin, ymax]"""
        if x_index is None: x_index = self._x_index
        if y_index is None: y_index = self._y_index

        if x_index != self.x_index or y_index != self.y_index or \
          self.axis.get_autoscalex_on() or self.axis.get_autoscaley_on():
            # One of them is not set or we work on a different set of axes
            T_lo, T_hi, P_lo, P_hi = self._get_Tp_limits()

            X = [0.0] * 4; Y = [0.0] * 4
            i = -1
            for T in [T_lo, T_hi]:
                for P in [P_lo, P_hi]:
                    i += 1
                    try:
                        self._state.update(CoolProp.PT_INPUTS, P, T)
                        # TODO: include a check for P and T?
                        X[i] = self._state.keyed_output(x_index)
                        Y[i] = self._state.keyed_output(y_index)
                    except:
                        X[i] = np.nan; Y[i] = np.nan
            # Figure out what to update
            dim = self._system[x_index]
            x_lim = [dim.from_SI(np.nanmin(X)), dim.from_SI(np.nanmax(X))]
            dim = self._system[y_index]
            y_lim = [dim.from_SI(np.nanmin(Y)), dim.from_SI(np.nanmax(Y))]
            # Either update the axes limits or get them
            if x_index == self._x_index:
                if self.axis.get_autoscalex_on():
                    self.axis.set_xlim(x_lim)
                else:
                    x_lim = self.axis.get_xlim()
            if y_index == self._y_index:
                if self.axis.get_autoscaley_on():
                    self.axis.set_ylim(y_lim)
                else:
                    y_lim = self.axis.get_ylim()
        else:  # We only asked for the real axes limits and they are set already
            x_lim = self.axis.get_xlim()
            y_lim = self.axis.get_ylim()

        return [x_lim[0], x_lim[1], y_lim[0], y_lim[1]]

    def _get_axis_limits(self, x_index=None, y_index=None):
        """Get the limits of the internal axis object in SI units
        Returns a list containing [xmin, xmax, ymin, ymax]"""
        if x_index is None: x_index = self._x_index
        if y_index is None: y_index = self._y_index
        limits = self.get_axis_limits(x_index, y_index)
        dim = self._system[x_index]
        limits[0] = dim.to_SI(limits[0])
        limits[1] = dim.to_SI(limits[1])
        dim = self._system[y_index]
        limits[2] = dim.to_SI(limits[2])
        limits[3] = dim.to_SI(limits[3])
        return limits

    @staticmethod
    def generate_ranges(itype, imin, imax, num):
        """Generate a range for a certain property"""
        if itype in [CoolProp.iP, CoolProp.iDmass]:
            return np.logspace(np.log2(imin), np.log2(imax), num=num, base=2.)
        return np.linspace(imin, imax, num=num)

    def _get_conversion_data(self):
        [Axmin, Axmax, Aymin, Aymax] = self._get_axis_limits()
        DELTAX_axis = Axmax - Axmin
        DELTAY_axis = Aymax - Aymin
        width = self.figure.get_figwidth()
        height = self.figure.get_figheight()
        pos = self.axis.get_position().get_points()
        [[Fxmin, Fymin], [Fxmax, Fymax]] = pos
        DELTAX_fig = width * (Fxmax - Fxmin)
        DELTAY_fig = height * (Fymax - Fymin)
        return [[Axmin, Axmax, Aymin, Aymax, Fxmin, Fxmax, Fymin, Fymax], [DELTAX_axis, DELTAY_axis, DELTAX_fig, DELTAY_fig]]

    def _to_pixel_coords(self, xv, yv):
        [[Axmin, Axmax, Aymin, Aymax, Fxmin, Fxmax, Fymin, Fymax], [DELTAX_axis, DELTAY_axis, DELTAX_fig, DELTAY_fig]] = self._get_conversion_data()
        # Convert coords to pixels
        x = (xv - Axmin) / DELTAX_axis * DELTAX_fig + Fxmin
        y = (yv - Aymin) / DELTAY_axis * DELTAY_fig + Fymin
        return x, y

    def _to_data_coords(self, xv, yv):
        [[Axmin, Axmax, Aymin, Aymax, Fxmin, Fxmax, Fymin, Fymax], [DELTAX_axis, DELTAY_axis, DELTAX_fig, DELTAY_fig]] = self._get_conversion_data()
        # Convert back to measurements
        x = (xv - Fxmin) / DELTAX_fig * DELTAX_axis + Axmin
        y = (yv - Fymin) / DELTAY_fig * DELTAY_axis + Aymin
        return x, y

    @staticmethod
    def get_x_y_dydx(xv, yv, x):
        """Get x and y coordinates and the linear interpolation derivative"""
        # Old implementation:
        # Get the rotation angle
        #f = interp1d(xv, yv)
        #y = f(x)
        #h = 0.00001*x
        #dy_dx = (f(x+h)-f(x-h))/(2*h)
        # return x,y,dy_dx
        if len(xv) == len(yv) and len(yv) > 1:  # assure same length
            if len(xv) == len(yv) and len(yv) == 2:  # only two points
                if np.min(xv) < x < np.max(xv):
                    dx = xv[1] - xv[0]
                    dy = yv[1] - yv[0]
                    dydx = dy / dx
                    y = yv[0] + dydx * (x - xv[0])
                    return x, y, dydx
                else:
                    raise ValueError("Your coordinate has to be between the input values.")
            else:
                limit = 1e-10                    # avoid hitting a point directly
                diff = np.array(xv) - x        # get differences
                index = np.argmin(diff * diff)  # nearest neighbour
                if (xv[index] < x < xv[index + 1]      # nearest below, positive inclination
                  or xv[index] > x > xv[index + 1]):   # nearest above, negative inclination
                    if diff[index] < limit:
                        index = [index - 1, index + 1]
                    else:
                        index = [index, index + 1]
                elif (xv[index - 1] < x < xv[index]    # nearest above, positive inclination
                  or xv[index - 1] > x > xv[index]):   # nearest below, negative inclination
                    if diff[index] < limit:
                        index = [index - 1, index + 1]
                    else:
                        index = [index - 1, index]
                xvnew = xv[index]
                yvnew = yv[index]
                return BasePlot.get_x_y_dydx(xvnew, yvnew, x)  # Allow for a single recursion
        else:
            raise ValueError("You have to provide the same amount of x- and y-pairs with at least two entries each.")

    def _inline_label(self, xv, yv, x=None, y=None):
        """
        This will give the coordinates and rotation required to align a label with
        a line on a plot in SI units.
        """
        if y is None and x is not None:
            trash = 0
            (xv, yv) = self._to_pixel_coords(xv, yv)
            # x is provided but y isn't
            (x, trash) = self._to_pixel_coords(x, trash)

            # Get the rotation angle and y-value
            x, y, dy_dx = BasePlot.get_x_y_dydx(xv, yv, x)
            rot = np.arctan(dy_dx) / np.pi * 180.

        elif x is None and y is not None:
            # y is provided, but x isn't
            _xv = xv[::-1]
            _yv = yv[::-1]
            # Find x by interpolation
            x = interpolate_values_1d(yv, xv, x_points=y)
            trash = 0
            (xv, yv) = self._to_pixel_coords(xv, yv)
            (x, trash) = self._to_pixel_coords(x, trash)

            # Get the rotation angle and y-value
            x, y, dy_dx = BasePlot.get_x_y_dydx(xv, yv, x)
            rot = np.arctan(dy_dx) / np.pi * 180.
        (x, y) = self._to_data_coords(x, y)
        return (x, y, rot)

    def inline_label(self, xv, yv, x=None, y=None):
        """
        This will give the coordinates and rotation required to align a label with
        a line on a plot in axis units.
        """
        dimx = self._system[self._x_index]
        xv = dimx.to_SI(xv)
        if x is not None: x = dimx.to_SI(x)
        dimy = self._system[self._y_index]
        yv = dimy.to_SI(yv)
        if y is not None: y = dimy.to_SI(y)
        (x, y, rot) = self._inline_label(xv, yv, x, y)
        x = dimx.from_SI(x)
        y = dimy.from_SI(y)
        return (x, y, rot)

    def show(self):
        plt.show()

    def savefig(self, *args, **kwargs):
        self.figure.savefig(*args, **kwargs)


if __name__ == "__main__":
    for sys in [SIunits(), KSIunits(), EURunits()]:
        print(sys.H.label)
        print(sys.H.to_SI(20))
        print(sys.P.label)
        print(sys.P.to_SI(20))

    # i_index, x_index, y_index, value=None, state=None)
    iso = IsoLine('T', 'H', 'P')
    print(iso.get_update_pair())

    state = AbstractState("HEOS", "water")
    iso = IsoLine('T', 'H', 'P', 300.0, state)
    hr = PropsSI("H", "T", [290, 310], "P", [1e5, 1e5], "water")
    pr = np.linspace(0.9e5, 1.1e5, 3)
    iso.calc_range(hr, pr)
    print(iso.x, iso.y)

    iso = IsoLine('Q', 'H', 'P', 0.0, state)
    iso.calc_range(hr, pr); print(iso.x, iso.y)
    iso = IsoLine('Q', 'H', 'P', 1.0, state)
    iso.calc_range(hr, pr); print(iso.x, iso.y)

    # bp = BasePlot(fluid_ref, graph_type, unit_system = 'KSI', **kwargs):
    bp = BasePlot('n-Pentane', 'PH', unit_system='EUR')
    # print(bp._get_sat_bounds('P'))
    # print(bp._get_iso_label(iso))
    print(bp.get_axis_limits())

        # get_update_pair(CoolProp.iP,CoolProp.iSmass,CoolProp.iT) -> (0,1,2,CoolProp.PSmass_INPUTS)
        # other values require switching and swapping
        # get_update_pair(CoolProp.iSmass,CoolProp.iP,CoolProp.iHmass) -> (1,0,2,CoolProp.PSmass_INPUTS)



from libcpp cimport bool 
from libcpp.string cimport string

# A header defining the AbstractState class
from . cimport cAbstractState

from .typedefs cimport *
from . cimport constants_header

#ctypedef fused string_t:
#    cython.p_char
#    bytes
#    unicode
#    string
#
#ctypedef fused string_or_size_t:
#    string_t
#    cython.integral

ctypedef fused string_or_size_t:
    cython.p_char
    bytes
    unicode
    string
    short
    int
    long

cdef class PyPhaseEnvelopeData:
    cpdef public bool TypeI
    cpdef public size_t iTsat_max, ipsat_max, icrit
    cpdef public list T, p, lnT, lnp, rhomolar_liq, rhomolar_vap, lnrhomolar_liq, lnrhomolar_vap, hmolar_liq, hmolar_vap, smolar_liq, smolar_vap, Q
    cpdef public list x, y, K
    
cdef class PyGuessesStructure:
    cpdef public double T, p, rhomolar, hmolar, smolar
    cpdef public double rhomolar_liq, rhomolar_vap
    cpdef public list x, y

cdef class PyCriticalState:
    cpdef public double T, p, rhomolar, hmolar, smolar
    cpdef public bool stable

cdef class PySpinodalData:
    cpdef public vector[double] tau, delta, M1

cdef class AbstractState:
    cdef cAbstractState.AbstractState *thisptr     # hold a C++ instance which we're wrapping
    cpdef update(self, constants_header.input_pairs iInput1, double Value1, double Value2)
    cpdef update_with_guesses(self, constants_header.input_pairs iInput1, double Value1, double Value2, PyGuessesStructure guesses)
    cpdef set_mole_fractions(self, vector[double] z)
    cpdef set_mass_fractions(self, vector[double] z)
    cpdef set_volu_fractions(self, vector[double] z)

    cpdef set_binary_interaction_double(self, string_or_size_t CAS1, string_or_size_t CAS2, string parameter, double val)
    cpdef double get_binary_interaction_double(self, string_or_size_t CAS1, string_or_size_t CAS2, string parameter) except *
    cpdef set_binary_interaction_string(self, string_or_size_t CAS1, string_or_size_t CAS2, string parameter, string val)
    cpdef string get_binary_interaction_string(self, string CAS1, string CAS2, string parameter) except *
    cpdef apply_simple_mixing_rule(self, size_t, size_t, string)
    
    cpdef name(self)
    cpdef backend_name(self)
    cpdef fluid_names(self)
    cpdef fluid_param_string(self, string key)
    cpdef set_fluid_parameter_double(self, size_t i, string parameter, double value)
    cpdef double get_fluid_parameter_double(self, size_t i, string parameter) except *
    cpdef change_EOS(self, size_t, string)
    
    cpdef constants_header.phases phase(self) except *  
    cpdef specify_phase(self, constants_header.phases phase)
    cpdef unspecify_phase(self)
    
    ## Limits
    cpdef double Tmin(self) except *
    cpdef double Tmax(self) except *
    cpdef double pmax(self) except *
    cpdef double Ttriple(self) except *
    
    ## Critical point
    cpdef double T_critical(self) except *
    cpdef double rhomass_critical(self) except *
    cpdef double rhomolar_critical(self) except *
    cpdef double p_critical(self) except *
    cpdef list all_critical_points(self)
    cpdef tuple criticality_contour_values(self)

    ## Spinodal curve(s)
    cpdef build_spinodal(self)
    cpdef PySpinodalData get_spinodal_data(self)
    
    ## Reducing point
    cpdef double T_reducing(self) except *
    cpdef double rhomolar_reducing(self) except *
    cpdef double rhomass_reducing(self) except *
    
    ## Tangent plane distance
    cpdef double tangent_plane_distance(self, double T, double p, vector[double] w, double rhomolar_guess=*) except *
        
    ## ---------------------------------------- 
    ##        Fluid property accessors
    ## ----------------------------------------
    
    cpdef double T(self) except *
    cpdef double p(self) except *
    cpdef double Q(self) except *
    cpdef double rhomolar(self) except *
    cpdef double hmolar(self) except *
    cpdef double smolar(self) except *
    cpdef double umolar(self) except *
    cpdef double cpmolar(self) except *
    cpdef double cp0molar(self) except *
    cpdef double cvmolar(self) except *
    cpdef double rhomass(self) except *
    cpdef double hmass(self) except *
    cpdef double smass(self) except *
    cpdef double umass(self) except *
    cpdef double cpmass(self) except *
    cpdef double cp0mass(self) except *
    cpdef double cvmass(self) except *
    cpdef double gibbsmass(self) except *
    cpdef double gibbsmolar(self) except *
    cpdef double helmholtzmass(self) except *
    cpdef double helmholtzmolar(self) except *
    cpdef double speed_sound(self) except *
    cpdef double gas_constant(self) except *
    cpdef double tau(self) except *
    cpdef double delta(self) except *
    cpdef double viscosity(self) except *
    cpdef double conductivity(self) except *
    cpdef dict conformal_state(self, string, CoolPropDbl, CoolPropDbl)
    cpdef dict conductivity_contributions(self)
    cpdef dict viscosity_contributions(self)
    cpdef double surface_tension(self) except *
    cpdef double Prandtl(self) except *
    cpdef double Bvirial(self) except *
    cpdef double Cvirial(self) except *
    cpdef double PIP(self) except *
    cpdef double fundamental_derivative_of_gas_dynamics(self) except *
    cpdef double isothermal_compressibility(self) except *
    cpdef double isobaric_expansion_coefficient(self) except *
    cpdef double fugacity(self, size_t) except *
    cpdef double fugacity_coefficient(self, size_t) except *
    cpdef double chemical_potential(self, size_t) except *

    cpdef double gibbsmolar_excess(self) except *
    cpdef double gibbsmass_excess(self) except *
    cpdef double hmolar_excess(self) except *
    cpdef double hmass_excess(self) except *
    cpdef double smolar_excess(self) except *
    cpdef double smass_excess(self) except *
    cpdef double umolar_excess(self) except *
    cpdef double umass_excess(self) except *
    cpdef double volumemolar_excess(self) except *
    cpdef double volumemass_excess(self) except *
    cpdef double helmholtzmolar_excess(self) except *
    cpdef double helmholtzmass_excess(self) except *

    


    
    cpdef double molar_mass(self) except *
    cpdef double acentric_factor(self) except*
    cpdef tuple true_critical_point(self)
    cpdef double get_fluid_constant(self,size_t,constants_header.parameters) except*
    cpdef double keyed_output(self, constants_header.parameters) except *
    cpdef double trivial_keyed_output(self, constants_header.parameters) except *
    cpdef double saturated_liquid_keyed_output(self, constants_header.parameters) except *
    cpdef double saturated_vapor_keyed_output(self, constants_header.parameters) except *
    
    cpdef tuple ideal_curve(self, string)
    
    ## ----------------------------------------	
    ##        Derivatives
    ## ----------------------------------------
    
    cpdef CoolPropDbl first_partial_deriv(self, constants_header.parameters, constants_header.parameters, constants_header.parameters) except *
    cpdef CoolPropDbl second_partial_deriv(self, constants_header.parameters, constants_header.parameters, constants_header.parameters, constants_header.parameters, constants_header.parameters) except *
    cpdef CoolPropDbl first_saturation_deriv(self, constants_header.parameters, constants_header.parameters) except *
    cpdef CoolPropDbl second_saturation_deriv(self, constants_header.parameters, constants_header.parameters, constants_header.parameters) except *
    
    cpdef double first_two_phase_deriv(self, constants_header.parameters Of, constants_header.parameters Wrt, constants_header.parameters Constant) except *
    cpdef double second_two_phase_deriv(self, constants_header.parameters Of, constants_header.parameters Wrt1, constants_header.parameters Constant1, constants_header.parameters Wrt2, constants_header.parameters Constant2) except *
    cpdef double first_two_phase_deriv_splined(self ,constants_header.parameters Of, constants_header.parameters Wrt, constants_header.parameters Constant, double x_end) except *
    
    cpdef double melting_line(self, int, int, double) except *
    cpdef bool has_melting_line(self) except *
    cpdef double saturation_ancillary(self, constants_header.parameters, int, constants_header.parameters, double) except *
        
    cpdef build_phase_envelope(self, string)
    cpdef PyPhaseEnvelopeData get_phase_envelope_data(self)
    
    cpdef mole_fractions_liquid(self)
    cpdef mole_fractions_vapor(self)
    cpdef get_mass_fractions(self) 
    cpdef get_mole_fractions(self)
    
    cpdef CoolPropDbl alpha0(self) except *
    cpdef CoolPropDbl dalpha0_dDelta(self) except *
    cpdef CoolPropDbl dalpha0_dTau(v) except *
    cpdef CoolPropDbl d2alpha0_dDelta2(self) except *
    cpdef CoolPropDbl d2alpha0_dDelta_dTau(self) except *
    cpdef CoolPropDbl d2alpha0_dTau2(self) except *
    cpdef CoolPropDbl d3alpha0_dTau3(self) except *
    cpdef CoolPropDbl d3alpha0_dDelta_dTau2(self) except *
    cpdef CoolPropDbl d3alpha0_dDelta2_dTau(self) except *
    cpdef CoolPropDbl d3alpha0_dDelta3(self) except *

    cpdef CoolPropDbl alphar(self) except *
    cpdef CoolPropDbl dalphar_dDelta(self) except *
    cpdef CoolPropDbl dalphar_dTau(self) except *
    cpdef CoolPropDbl d2alphar_dDelta2(self) except *
    cpdef CoolPropDbl d2alphar_dDelta_dTau(self) except *
    cpdef CoolPropDbl d2alphar_dTau2(self) except *
    cpdef CoolPropDbl d3alphar_dDelta3(self) except *
    cpdef CoolPropDbl d3alphar_dDelta2_dTau(self) except *
    cpdef CoolPropDbl d3alphar_dDelta_dTau2(self) except *
    cpdef CoolPropDbl d3alphar_dTau3(self) except *
    cpdef CoolPropDbl d4alphar_dDelta4(self) except *
    cpdef CoolPropDbl d4alphar_dDelta3_dTau(self) except *
    cpdef CoolPropDbl d4alphar_dDelta2_dTau2(self) except *
    cpdef CoolPropDbl d4alphar_dDelta_dTau3(self) except *
    cpdef CoolPropDbl d4alphar_dTau4(self) except *




import unittest
from CoolProp.CoolProp import PropsSI
import CoolProp
import numpy as np


def test_input_types():
    for Fluid in ['Water']:
        for Tvals in [0.5 * PropsSI(Fluid, 'Tmin') + 0.5 * PropsSI(Fluid, 'Tcrit'),
                      [PropsSI(Fluid, 'Tmin') + 1e-5, PropsSI(Fluid, 'Tcrit') - 1e-5],
                      np.linspace(PropsSI(Fluid, 'Tmin') + 1e-5, PropsSI(Fluid, 'Tcrit') - 1e-5, 30)
                      ]:
            yield check_type, Fluid, Tvals


def check_type(fluid, Tvals):
    PropsSI('P', 'T', Tvals, 'Q', 0, fluid)


class PropsFailures(unittest.TestCase):
    def testUnmatchedLengths(self):
        self.assertRaises(TypeError, PropsSI, 'P', 'T', [280, 290, 300], 'Q', [0, 1], 'R134a')

    def testMatrix(self):
        self.assertRaises(TypeError, PropsSI, 'P', 'T', np.array([280, 290, 300, 280, 290, 300]).reshape(2, 3), 'Q', np.array([0, 0.5, 1, 0.0, 0.5, 1]).reshape(2, 3), 'R134a')


if __name__ == '__main__':
    import nose
    nose.runmodule()
